'''
Created on 2016/5/30

:author: hubo
'''


from vlcp.config.config import defaultconfig
from vlcp.server.module import depend, publicapi, callAPI
from vlcp.service.sdn.flowbase import FlowBase
from vlcp.service.sdn.ofpmanager import FlowInitialize
from vlcp.utils.flowupdater import FlowUpdater
from vlcp.utils.networkmodel import PhysicalPort, LogicalPort, LogicalNetwork, VXLANEndpointSet,\
    LogicalNetworkMap, LogicalPortVXLANInfo
import vlcp.service.kvdb.objectdb as objectdb
import vlcp.service.sdn.ofpportmanager as ofpportmanager
from vlcp.event.runnable import RoutineContainer
from vlcp.protocol.openflow.openflow import OpenflowConnectionStateEvent,\
    OpenflowAsyncMessageEvent, OpenflowErrorResultException
from vlcp.utils.ethernet import ethernet_l2, mac_addr, mac_addr_bytes
import vlcp.service.sdn.ioprocessing as iop
import itertools
from namedstruct.namedstruct import dump
from vlcp.event.event import Event, withIndices
from vlcp.service.sdn import ovsdbportmanager
from vlcp.utils import ovsdb
from vlcp.utils.dataobject import updater, ReferenceObject
from time import time
import vlcp.utils.vxlandiscover as vxlandiscover
from vlcp.utils.vxlandiscover import get_broadcast_ips
import json
try:
    reduce
except Exception:
    from functools import reduce


@withIndices('connection', 'logicalnetworkid', 'type')
class VXLANGroupChanged(Event):
    UPDATED = 'updated'
    DELETED = 'deleted'

def _is_vxlan(obj):
    try:
        return obj.physicalnetwork.type == 'vxlan'
    except AttributeError:
        return False

def _get_ip(ip, ofdef):
    try:
        return ofdef.ip4_addr(ip)
    except Exception:
        return None

class VXLANUpdater(FlowUpdater):
    def __init__(self, connection, parent):
        FlowUpdater.__init__(self, connection, (), ('VXLANUpdater', connection), parent._logger)
        self._parent = parent
        self._lastlognets = ()
        self._lastphyports = ()
        self._lastlogports = ()
        self._lastphyportinfo = {}
        self._lastlognetinfo = {}
        self._lastlogportinfo = {}
        self._lastvxlaninfo = {}
        self._orig_initialkeys = ()
        self._watched_maps = []
        # For first-upload mode, save the watching MAC addresses
        # a dictionary (logicalnetwork, mac_address) -> expire time
        self._watched_macs = {}
        # Current buffered packets, a dictionary (logicalnetwork, mac_address) -> [packet, packet, ...]
        # where packet is (buffer_id, data, expire, in_port, metadata)
        self._buffered_packets = {}
        # LogicalNetworkID -> PhysicalPortNo map
        self._current_groups = {}
        self._walk_lognet = vxlandiscover.lognet_vxlan_walker(self._parent.prepush)
    def wait_for_group(self, container, networkid, timeout = 30):
        if networkid in self._current_groups:
            container.retvalue = self._current_groups[networkid]
        else:
            groupchanged = VXLANGroupChanged.createMatcher(self._connection, networkid, VXLANGroupChanged.UPDATED)
            for m in container.waitWithTimeout(timeout, groupchanged):
                yield m
            if container.timeout:
                raise ValueError('VXLAN group is still not created after a long time')
            else:
                container.retvalue = container.event.physicalportid
    def main(self):
        try:
            if self._connection.protocol.disablenxext:
                return
            self.subroutine(self._update_handler(), name = '_update_handler_routine')
            self.subroutine(self._refresh_handler(), name = '_refresh_handler_routine')
            if not self._parent.prepush and not self._parent.learning:
                self.subroutine(self._query_packet_handler(), name = '_query_packet_handler_routine')
            for m in FlowUpdater.main(self):
                yield m
        finally:
            if hasattr(self, '_update_handler_routine'):
                self._update_handler_routine.close()
            if hasattr(self, '_refresh_handler_routine'):
                self._refresh_handler_routine.close()
            if hasattr(self, '_query_packet_handler_routine'):
                self._query_packet_handler_routine.close()                           
    def reset_initialkeys(self, keys, values):
        if self._parent.prepush:
            self._watched_maps = [k for k,v in zip(keys, values)
                                  if v is not None and not v.isdeleted() and v.isinstance(LogicalNetworkMap)]
            # If the logical network map changed, restart the walk process
            self._initialkeys = tuple(itertools.chain(self._orig_initialkeys, self._watched_maps))
    def _walk_phyport(self, key, value, walk, save):
        save(key)
    def _walk_logport(self, key, value, walk, save):
        save(key)
    def _walk_mac_key(self, key, value, walk, save):
        save(key)
        if value is not None:
            logport_key = value.ref.getkey()
            try:
                _ = walk(logport_key)
            except KeyError:
                pass
            else:
                save(logport_key)
            _, (portid,) = LogicalPort._getIndices(logport_key)
            try:
                portinfokey = LogicalPortVXLANInfo.default_key(portid)
                _ = walk(portinfokey)
            except KeyError:
                pass
            else:
                save(portinfokey)
    def _get_watched_mac_keys(self):
        keys = []
        current_time = time()
        changed = False
        for k,t in list(self._watched_macs.items()):
            if t is not None and t < current_time:
                # Expired
                del self._watched_macs[k]
                changed = True
                if k in self._buffered_packets:
                    # The packets should be expired far ago
                    del self._buffered_packets[k]
            elif k[0] not in (n for n,_ in self._lastlognets):
                # Network is removed
                del self._watched_macs[k]
                changed = True
                if k in self._buffered_packets:
                    del self._buffered_packets[k]
            else:
                keys.append(LogicalPort.unique_key('_mac_address_index', *k))
        return (keys, changed)
    def _update_handler(self):
        dataobjectchanged = iop.DataObjectChanged.createMatcher(None, None, self._connection)
        while True:
            yield (dataobjectchanged,)
            self._lastlogports, self._lastphyports, self._lastlognets, _ = self.event.current
            self._update_walk()
            self.updateobjects([p for p,_ in self._lastlogports])
    def _update_walk(self):
        phyport_keys = [p.getkey() for p,_ in self._lastphyports]
        lognet_keys = [n.getkey() for n,_ in self._lastlognets]
        logport_keys = [p.getkey() for p,_ in self._lastlogports]
        mac_keys, _ = self._get_watched_mac_keys()
        self._orig_initialkeys = phyport_keys + lognet_keys + logport_keys + mac_keys
        self._initialkeys = self._orig_initialkeys + self._watched_maps
        self._walkerdict = dict(itertools.chain(((n, self._walk_lognet) for n in lognet_keys),
                                                ((p, self._walk_phyport) for p in phyport_keys),
                                                ((p, self._walk_logport) for p in logport_keys),
                                                ((p, self._walk_mac_key) for p in mac_keys)))
        self.subroutine(self.restart_walk(), False)
    def _query_packet_handler(self):
        ofdef = self._connection.openflowdef
        vhost = self._connection.protocol.vhost
        vo = self._parent._gettableindex('vxlanoutput', vhost)
        packet_in = OpenflowAsyncMessageEvent.createMatcher(ofdef.OFPT_PACKET_IN, None, None,
                                                            vo, 2, self._connection, self._connection.connmark)
        flow_remove = OpenflowAsyncMessageEvent.createMatcher(ofdef.OFPT_FLOW_REMOVED, None, None,
                                                              vo, 3, self._connection, self._connection.connmark,
                                                              _ismatch = lambda x: x.message.reason == ofdef.OFPRR_IDLE_TIMEOUT)
        while True:
            for m in self.waitWithTimeout(self._parent.pushtimeout, packet_in, flow_remove):
                yield m
            if self.timeout:
                _, changed = self._get_watched_mac_keys()
                if changed:
                    self._update_walk()
            elif self.matcher is packet_in:
                msg = self.event.message
                try:
                    packet = ethernet_l2.create(msg.data)
                except Exception:
                    self._logger.warning('Invalid packet received: %r', msg.data, exc_info = True)
                else:
                    in_port = ofdef.ofp_port_no.create(ofdef.get_oxm(msg.match.oxm_fields, ofdef.OXM_OF_IN_PORT))
                    metadata = [o for o in msg.match.oxm_fields if o.header in (ofdef.NXM_NX_REG4,
                                                                                   ofdef.NXM_NX_REG5,
                                                                                   ofdef.NXM_NX_REG6)]
                    mac_address = mac_addr.formatter(packet.dl_dst)
                    nid = ofdef.uint32.create(ofdef.get_oxm(msg.match.oxm_fields, ofdef.NXM_NX_REG5))
                    networks = [n for n,nid2 in self._lastlognets if nid2 == nid]
                    if networks:
                        network = networks[0]
                        current_time = time()
                        if (network, mac_address) not in self._watched_macs:
                            self._watched_macs[(network, mac_address)] = current_time + self._parent.pushtimeout
                            self._buffered_packets[(network, mac_address)] = [(msg.buffer_id,
                                                                               msg.data if msg.buffer_id == ofdef.OFP_NO_BUFFER
                                                                               else b'',
                                                                               current_time + self._parent.packetexpire,
                                                                               in_port,
                                                                               metadata
                                                                               )]
                            # Restart walk since we have already changed the watch list
                            self._update_walk()
                        else:
                            if self._watched_macs[(network, mac_address)] is not None:
                                # Update expire time
                                self._watched_macs[(network, mac_address)] = current_time + self._parent.pushtimeout
                            # If the buffers are already removed, the flow is already sent. Ignore this packet.
                            # This should rarely happen, so we do not bother drop it with a PACKET_OUT message
                            if (network, mac_address) in self._buffered_packets:
                                l = self._buffered_packets[(network, mac_address)]
                                # Drop expired packets
                                l = [p for p in l if p[2] > current_time]
                                l.append((msg.buffer_id,
                                           msg.data if msg.buffer_id == ofdef.OFP_NO_BUFFER
                                           else b'',
                                           current_time + self._parent.packetexpire,
                                           in_port,
                                           metadata
                                           ))
                                self._buffered_packets[(network, mac_address)] = l
            else:
                # A flow is expired, also remove the watch list
                msg = self.event.message
                nid = ofdef.uint32.create(ofdef.get_oxm(msg.match.oxm_fields, ofdef.NXM_NX_REG5))
                mac_address = mac_addr_bytes.formatter(ofdef.get_oxm(msg.match.oxm_fields, ofdef.OXM_OF_ETH_DST))
                networks = [n for n,nid2 in self._lastlognets if nid2 == nid]
                if networks:
                    network = networks[0]
                    if (network, mac_address) in self._watched_macs:
                        del self._watched_macs[(network, mac_address)]
                        if (network, mac_address) in self._buffered_packets:
                            del self._buffered_packets[(network, mac_address)]
                        self._update_walk()
    def _refresh_handler(self):
        while True:
            for m in self.waitWithTimeout(self._parent.refreshinterval):
                yield m
            self.updateobjects(n for n,_ in self._lastlognets)
    def updateflow(self, conn, addvalues, removevalues, updatedvalues):
        # Following works are done in parallel:
        # 1. Modify the VXLANEndpointSet in KVDB
        # 2. Create/Modify the broadcast group for every logical network in VXLAN
        # 3. Create MAC-address <-> Tunnel IP address map in KVDB
        # 4. Create/Modify learning flows, or push the flows for every logical port
        ofdef = conn.openflowdef
        vhost = conn.protocol.vhost
        vi = self._parent._gettableindex('vxlaninput', vhost)
        vi_next = self._parent._getnexttable('', 'vxlaninput', vhost = vhost)
        vo = self._parent._gettableindex('vxlanoutput', vhost)
        vo_next = self._parent._getnexttable('', 'vxlanoutput', vhost = vhost)
        learning = self._parent._gettableindex('vxlanlearning', vhost)
        egress = self._parent._gettableindex('egress', vhost)
        allresult = set(v for v in self._savedresult if v is not None and not v.isdeleted())
        lastlognetinfo = self._lastlognetinfo
        lastphyportinfo = self._lastphyportinfo
        lastlogportinfo = self._lastlogportinfo
        currentlognetinfo = dict((n,(n.physicalnetwork,nid)) for n,nid in self._lastlognets if n in allresult and _is_vxlan(n))
        currentlogportinfo = dict((p, (p.network, pid, getattr(p, 'mac_address', None))) for p,pid in self._lastlogports
                                  if p.network in currentlognetinfo)
        phyportdict = dict((p,pid) for p,pid in self._lastphyports if p in allresult and _is_vxlan(p))
        # We only accept one physical port for VXLAN network
        unique_phyports = dict((p.physicalnetwork, p) for p,v in sorted(phyportdict.items(), key = lambda x: x[1]))
        phyportdict = dict((p,phyportdict[p]) for p in unique_phyports.values())
        currentphyportinfo = dict((p,(p.physicalnetwork,pid,lastphyportinfo[p][2])) for p,pid in phyportdict.items()
                                  if p in lastphyportinfo and lastphyportinfo[p][1] == pid)
        newports = set(phyportdict.keys())
        newports.difference_update(currentphyportinfo.keys())
        removed_ports = set(lastphyportinfo.keys()).difference(currentphyportinfo.keys())
        # Retrieve VXLAN settings from OVSDB
        newphyports = []
        datapath_id = conn.openflow_datapathid
        ovsdb_vhost = self._parent.vhostmap.get(vhost, "")
        try:
            for m in callAPI(self, 'ovsdbmanager', 'waitbridgeinfo', {'datapathid' : datapath_id,
                                                                     'vhost' : ovsdb_vhost}):
                yield m
        except Exception:
            self._logger.warning("OVSDB bridge is not ready", exc_info = True)
            return
        else:
            bridge, system_id, _ = self.retvalue
        if newports:
            try:
                for m in callAPI(self, 'ovsdbmanager', 'waitconnection', {'datapathid': datapath_id,
                                                                         'vhost': ovsdb_vhost}):
                    yield m
            except Exception:
                self._parent._logger.warning("OVSDB connection is not ready for datapathid = %016x, vhost = %r(%r)",
                                             datapath_id, vhost, ovsdb_vhost, exc_info = True)
            else:
                try:
                    ovsdb_conn = self.retvalue
                    port_requests = [(p, phyportdict[p]) for p in newports]
                    def wait_and_ignore(portno):
                        try:
                            for m in callAPI(self, 'ovsdbportmanager', 'waitportbyno', {'datapathid': datapath_id,
                                                                                 'portno': portno,
                                                                                 'timeout': 1,
                                                                                 'vhost': ovsdb_vhost}):
                                yield m
                        except ovsdbportmanager.OVSDBPortNotAppearException:
                            self.retvalue = None
                    for m in self.executeAll([wait_and_ignore(pid) for _,pid in port_requests]):
                        yield m
                    uuids = [r[0]['_uuid'] for r in self.retvalue]
                    method, params = ovsdb.transact('Open_vSwitch',
                                                    *[ovsdb.select('Interface', [["_uuid", "==", ovsdb.uuid(u)]],
                                                                                 ["_uuid", "options"])
                                                      for u in uuids])
                    for m in ovsdb_conn.protocol.querywithreply(method, params, ovsdb_conn, self):
                        yield m
                    src_ips = [ovsdb.omap_getvalue(r['rows'][0]['options'], "local_ip")
                               if 'error' not in r and r['rows'] and 'options' in r['rows'][0]
                               else None
                               for r in self.jsonrpc_result]
                    newphyports = [(p, (p.physicalnetwork, pid, src_ip))
                                  for (p, pid),src_ip in zip(port_requests, src_ips)
                                  if _get_ip(src_ip, ofdef) is not None]
                    currentphyportinfo.update(newphyports)
                except Exception:
                    self._parent._logger.warning("Get VXLAN configurations from OVSDB failed for datapathid = %016x, vhost = %r(%r)",
                                                 datapath_id, vhost, ovsdb_vhost, exc_info = True)
        # If there are any added or removed physical ports, do transact on all logical networks;
        # Do transact on added and removed logical networks else
        transact_networks = set(currentlognetinfo.keys()).symmetric_difference(lastlognetinfo.keys())
        if newphyports:
            newphynets = set(v[0] for _,v in newphyports)
            transact_networks.update(lognet for lognet, (phynet, _) in currentlognetinfo.items()
                                     if phynet in newphynets)
        if removed_ports:
            removephynets = set(p.physicalnetwork for p in removed_ports)
            transact_networks.update(lognet for lognet, (phynet, _) in lastlognetinfo.items()
                                     if phynet in removephynets)
        transact_networks.update(n for n in updatedvalues if n.isinstance(LogicalNetwork) and n in currentlognetinfo)
        # LogicalPortVXLANInfo
        lastvxlaninfo = self._lastvxlaninfo
        otherlogports = dict((p.id, p) for p in allresult
                           if p.isinstance(LogicalPort) and hasattr(p, 'mac_address') and p not in currentlogportinfo)
        currentvxlaninfo = dict((obj, (otherlogports[obj.id], otherlogports[obj.id].network,
                                       currentlognetinfo[otherlogports[obj.id].network][1],
                                       otherlogports[obj.id].network.vni,
                                       otherlogports[obj.id].mac_address,
                                       obj.endpoints[0],
                                       unique_phyports[otherlogports[obj.id].network.physicalnetwork],
                                       currentphyportinfo[unique_phyports[otherlogports[obj.id].network.physicalnetwork]][1]))
                                for obj in allresult
                                if obj.isinstance(LogicalPortVXLANInfo) and obj.id in otherlogports \
                                    and hasattr(obj, 'endpoints') and obj.endpoints \
                                    and (obj.endpoints[0]['vhost'], obj.endpoints[0]['systemid'], obj.endpoints[0]['bridge'])
                                            != (ovsdb_vhost, system_id, bridge) \
                                    and hasattr(otherlogports[obj.id].network, 'vni') \
                                    and otherlogports[obj.id].network in currentlognetinfo
                                    and otherlogports[obj.id].network.physicalnetwork in unique_phyports)
        #=======================================================================
        # self._parent._logger.debug("VXLAN info updated:\n%s\n\nLast VXLAN info:\n%s",
        #                            '\n'.join("%s (%s, %r) -> %s" % (k.id, v[4], v[3], v[5]['tunnel_dst'])
        #                                      for k,v in currentvxlaninfo.items()),
        #                            '\n'.join("%s (%s, %r) -> %s" % (k.id, v[4], v[3], v[5]['tunnel_dst'])
        #                                      for k,v in lastvxlaninfo.items()))
        #=======================================================================
        self._lastvxlaninfo = currentvxlaninfo
        self._lastphyportinfo = currentphyportinfo
        self._lastlognetinfo = currentlognetinfo
        self._lastlogportinfo = currentlogportinfo
        subroutines = []
        removed_ports = {}
        created_ports = {}
        def _remove_port_tun(lognet, logport):
            if lognet in lastlognetinfo:
                phynet, _ = lastlognetinfo[lognet]
                phyport = [(p,v) for p,v in lastphyportinfo.items() if v[0] == phynet]
                if phyport:
                    phyport = phyport[0]
                    removed_ports[logport.id] = phyport[1][2]
        def _create_port_tun(lognet, logport):
            if lognet in currentlognetinfo:
                phynet, _ = currentlognetinfo[lognet]
                phyport = [(p,v) for p,v in currentphyportinfo.items() if v[0] == phynet]
                if phyport:
                    phyport = phyport[0]
                    created_ports[logport.id] = phyport[1][2]
        for logport, (lognet, _, _) in lastlogportinfo.items():
            if lognet in transact_networks:
                _remove_port_tun(lognet, logport)
        for logport in lastlogportinfo:
            if logport not in currentlogportinfo or currentlogportinfo[logport] != lastlogportinfo[logport]:
                lognet, _, _ = lastlogportinfo[logport]
                _remove_port_tun(lognet, logport)
        for logport in currentlogportinfo:
            if logport not in lastlogportinfo or lastlogportinfo[logport] != currentlogportinfo[logport]:
                lognet, _, _ = currentlogportinfo[logport]
                _create_port_tun(lognet, logport)
        for logport, (lognet, _, _) in currentlogportinfo.items():
            if lognet in transact_networks:
                _create_port_tun(lognet, logport)
        if transact_networks or created_ports or removed_ports:
            # Network -> IP dictionary
            network_ip_dict = {}
            for n in transact_networks:
                if n is not None and not n.isdeleted() and n.physicalnetwork in unique_phyports:
                    phyport = unique_phyports[n.physicalnetwork]
                    if phyport in currentphyportinfo:
                        network_ip_dict[n.id] = currentphyportinfo[phyport][2]
                    else:
                        network_ip_dict[n.id] = None
                else:
                    network_ip_dict[n.id] = None
            subroutines.append(vxlandiscover.update_vxlaninfo(self, network_ip_dict,
                                                              created_ports, removed_ports,
                                                              ovsdb_vhost, system_id, bridge,
                                                              self._parent.allowedmigrationtime,
                                                              self._parent.refreshinterval))
        # We must create broadcast groups for VXLAN logical networks, using the information in VXLANEndpointSet
        group_cmds = []
        created_groups = {}
        deleted_groups = set()
        for lognet in removevalues:
            if lognet.isinstance(LogicalNetwork):
                if lognet in lastlognetinfo:
                    phynet, netid = lastlognetinfo[lognet]
                    if netid in self._current_groups:
                        deleted_groups.add(netid)
                        del self._current_groups[netid]
                    # Always delete to ensure non-exists
                    group_cmds.append(ofdef.ofp_group_mod(command = ofdef.OFPGC_DELETE,
                                                          type = ofdef.OFPGT_ALL,
                                                          group_id = (netid & 0xffff) | 0x10000))
                    group_cmds.append(ofdef.ofp_group_mod(command = ofdef.OFPGC_DELETE,
                                                          type = ofdef.OFPGT_ALL,
                                                          group_id = (netid & 0xffff) | 0x20000))
        for ve in itertools.chain(addvalues, updatedvalues):
            if ve.isinstance(VXLANEndpointSet):
                lognet = ReferenceObject(LogicalNetwork.default_key(ve.id))
                if lognet in currentlognetinfo:
                    phynet, netid = currentlognetinfo[lognet]
                    if phynet in unique_phyports:
                        phyport = unique_phyports[phynet]
                        if phyport in currentphyportinfo:
                            _, portid, localip = currentphyportinfo[phyport]
                            allips = [ipnum for _,ipnum in get_broadcast_ips(ve, [localip],
                                                                             ovsdb_vhost, system_id, bridge)]
                            created_groups[netid] = portid
                            group_cmds.append(
                                ofdef.ofp_group_mod(
                                      command = ofdef.OFPGC_MODIFY
                                                if netid in self._current_groups
                                                else ofdef.OFPGC_ADD,
                                      type = ofdef.OFPGT_ALL,
                                      group_id = (netid & 0xffff) | 0x10000,
                                      buckets =
                                        [ofdef.ofp_bucket(
                                                actions = [
                                                    ofdef.ofp_action_set_field(
                                                        field = ofdef.create_oxm(
                                                                    ofdef.NXM_NX_TUN_IPV4_DST,
                                                                    ipaddress
                                                                )
                                                    ),
                                                    ofdef.ofp_action_output(port = portid)
                                                    ]
                                            )
                                        for ipaddress in allips
                                        ]
                                ))
                            group_cmds.append(
                                ofdef.ofp_group_mod(
                                      command = ofdef.OFPGC_MODIFY
                                                if netid in self._current_groups
                                                else ofdef.OFPGC_ADD,
                                      type = ofdef.OFPGT_ALL,
                                      group_id = (netid & 0xffff) | 0x20000,
                                      buckets =
                                        [ofdef.ofp_bucket(
                                                actions = [
                                                    ofdef.ofp_action_set_field(
                                                        field = ofdef.create_oxm(
                                                                    ofdef.NXM_NX_TUN_IPV4_DST,
                                                                    ipaddress
                                                                )
                                                    ),
                                                    ofdef.ofp_action_output(port = ofdef.OFPP_IN_PORT)
                                                    ]
                                            )
                                        for ipaddress in allips
                                        ]
                                ))
                    else:
                        created_groups[netid] = None
                        group_cmds.append(
                            ofdef.ofp_group_mod(
                                  command = ofdef.OFPGC_MODIFY
                                            if netid in self._current_groups
                                            else ofdef.OFPGC_ADD,
                                  type = ofdef.OFPGT_ALL,
                                  group_id = (netid & 0xffff) | 0x10000,
                                  buckets = []
                            ))
                        group_cmds.append(
                            ofdef.ofp_group_mod(
                                  command = ofdef.OFPGC_MODIFY
                                            if netid in self._current_groups
                                            else ofdef.OFPGC_ADD,
                                  type = ofdef.OFPGT_ALL,
                                  group_id = (netid & 0xffff) | 0x20000,
                                  buckets = []
                            ))                        
        if group_cmds:
            def group_mod():
                try:
                    for m in conn.protocol.batch(group_cmds, conn, self):
                        yield m
                except Exception:
                    self._parent._logger.warning("Some Openflow commands return error result on connection %r, will ignore and continue.\n"
                                                 "Details:\n%s", conn,
                                                 "\n".join("REQUEST = \n%s\nERRORS = \n%s\n" % (json.dumps(dump(k), indent=2),
                                                                                                json.dumps(dump(v), indent=2))
                                                           for k,v in self.openflow_replydict.items()))
                # Some groups are not modified, but the physical port may change, we should also send VXLANGroupChanged
                for lognet in currentlognetinfo:
                    phynet, netid = currentlognetinfo[lognet]
                    if netid in self._current_groups and netid not in created_groups:
                        if phynet in unique_phyports:
                            phyport = unique_phyports[phynet]
                            if phyport in currentphyportinfo:
                                _, portid, _ = currentphyportinfo[phyport]
                                if portid != self._current_groups[netid]:
                                    created_groups[netid] = portid
                self._current_groups.update(created_groups)
                for g in deleted_groups:
                    for m in self.waitForSend(VXLANGroupChanged(conn, g, VXLANGroupChanged.DELETED)):
                        yield m
                for g,pid in created_groups.items():
                    for m in self.waitForSend(VXLANGroupChanged(conn, g, VXLANGroupChanged.UPDATED, physicalportid = pid)):
                        yield m
            subroutines.append(group_mod())
        # Finally, lets create flow entries for VXLAN input/output
        # There are three patterns with different options:
        # 1. learning: use nx_action_learn action for learning
        # 2. prepush: for each LogicalPortVXLANInfo, send a flow entry to set the tunnel destination
        # 3. prepush = False and learning = False: use output(CONTROLLER) to query at the first packet
        # We always require nicira extension, so we are not using output(CONTROLLER) for learning.
        if conn.protocol.disablenxext:
            def _create_oxms(*fields):
                return [ofdef.create_oxm(ofdef.OXM_OF_METADATA_W, reduce(lambda x,y: x | y, [f[0] for f in fields]),
                                                                reduce(lambda x,y: x | y, [f[1] for f in fields]))]
            def _outport_oxm(pid):
                return (pid & 0xffff, 0xffff)
            def _innet_oxm(nid):
                return ((nid & 0xffff) << 48, 0xffff000000000000)
            def _outnet_oxm(nid):
                return ((nid & 0xffff) << 32, 0xffff00000000)
            def _flags(learned, in_port):
                return ((int(bool(learned)) << 16) | (int(bool(in_port)) << 31), 0x80010000)
        else:
            def _create_oxms(*fields):
                return list(fields)
            def _outport_oxm(pid):
                return ofdef.create_oxm(ofdef.NXM_NX_REG6, pid)
            def _innet_oxm(nid):
                return ofdef.create_oxm(ofdef.NXM_NX_REG4, nid)
            def _outnet_oxm(nid):
                return ofdef.create_oxm(ofdef.NXM_NX_REG5, nid)
            def _flags(learned, in_port):
                return ofdef.create_oxm(ofdef.NXM_NX_REG7_W, int(bool(learned)) | ((int(bool(in_port)) << 15)), 0x8001)
        def flow_mod():
            cmds = []
            for p in itertools.chain(removevalues, updatedvalues):
                if p.isinstance(PhysicalPort) and p in lastphyportinfo and not p in currentphyportinfo:
                    _, pid, _ = lastphyportinfo[p]
                    if self._parent.learning:
                        cmds.append(ofdef.ofp_flow_mod(table_id = vi,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(
                                                                    oxm_fields = [
                                                                        ofdef.create_oxm(ofdef.OXM_OF_IN_PORT, pid)
                                                                        ]
                                                                )))
                    cmds.append(ofdef.ofp_flow_mod(table_id = vo,
                                                   command = ofdef.OFPFC_DELETE_STRICT,
                                                   priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                   out_port = ofdef.OFPP_ANY,
                                                   out_group = ofdef.OFPG_ANY,
                                                   match = ofdef.ofp_match_oxm(
                                                                oxm_fields = _create_oxms(
                                                                        _outport_oxm(pid)
                                                                    )
                                                            )))
                    cmds.append(ofdef.ofp_flow_mod(table_id = egress,
                                                   command = ofdef.OFPFC_DELETE_STRICT,
                                                   priority = ofdef.OFP_DEFAULT_PRIORITY + 10,
                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                   out_port = ofdef.OFPP_ANY,
                                                   out_group = ofdef.OFPG_ANY,
                                                   match = ofdef.ofp_match_oxm(
                                                                oxm_fields = _create_oxms(
                                                                    _outport_oxm(pid),
                                                                    _flags(True, False)
                                                                    )
                                                            )))
                    cmds.append(ofdef.ofp_flow_mod(table_id = egress,
                                                   command = ofdef.OFPFC_DELETE_STRICT,
                                                   priority = ofdef.OFP_DEFAULT_PRIORITY + 10,
                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                   out_port = ofdef.OFPP_ANY,
                                                   out_group = ofdef.OFPG_ANY,
                                                   match = ofdef.ofp_match_oxm(
                                                                oxm_fields = _create_oxms(
                                                                    _outport_oxm(pid),
                                                                    _flags(True, True)
                                                                    ) + [ofdef.create_oxm(ofdef.OXM_OF_IN_PORT, pid)]
                                                            )))
            for m in self.execute_commands(conn, cmds):
                yield m
            del cmds[:]
            if self._parent.learning:
                # Also delete all learned flows
                for p in itertools.chain(removevalues, updatedvalues):
                    if p.isinstance(PhysicalPort) and p in lastphyportinfo:
                        _, pid, _ = lastphyportinfo[p]
                        cmds.append(ofdef.ofp_flow_mod(table_id = learning,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(
                                                                    oxm_fields = _create_oxms(
                                                                        _outport_oxm(pid)
                                                                        )
                                                                )))
                    elif p.isinstance(LogicalNetwork) and p in lastlognetinfo and not p in currentlognetinfo:
                        _, nid = lastlognetinfo[p]
                        cmds.append(ofdef.ofp_flow_mod(table_id = learning,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(
                                                                    oxm_fields = _create_oxms(
                                                                        _outnet_oxm(nid)
                                                                        )
                                                                )))
                for m in self.execute_commands(conn, cmds):
                    yield m
            for p in itertools.chain(addvalues, updatedvalues):
                if p.isinstance(PhysicalPort) and p in currentphyportinfo and p not in lastphyportinfo:
                    _, pid, _ = currentphyportinfo[p]
                    if self._parent.learning:
                        cmds.append(ofdef.ofp_flow_mod(table_id = vi,
                                                       command = ofdef.OFPFC_ADD,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(
                                                                    oxm_fields = [
                                                                        ofdef.create_oxm(ofdef.OXM_OF_IN_PORT, pid)
                                                                        ]
                                                                ),
                                                       instructions = [
                                                                ofdef.ofp_instruction_actions(
                                                                    actions = [
                                                                            ofdef.nx_action_learn(
                                                                                hard_timeout = self._parent.learntimeout,
                                                                                priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                                cookie = 0,
                                                                                table_id = learning,
                                                                                specs = [ofdef.create_nxfms_matchvalue(ofdef.NXM_NX_REG6, pid),                 # Match port
                                                                                         ofdef.create_nxfms_matchfield(ofdef.OXM_OF_ETH_SRC, ofdef.OXM_OF_ETH_DST), # Match dst MAC adddress = src MAC address
                                                                                         ofdef.create_nxfms_matchfield(ofdef.NXM_NX_REG4, ofdef.NXM_NX_REG5),       # Match output network = input network
                                                                                         ofdef.create_nxfms_loadfield(ofdef.OXM_OF_TUNNEL_ID, ofdef.OXM_OF_TUNNEL_ID),   # Set tunnel ID
                                                                                         ofdef.create_nxfms_loadfield(ofdef.NXM_NX_TUN_IPV4_SRC, ofdef.NXM_NX_TUN_IPV4_DST),
                                                                                         ofdef.create_nxfms_loadvalue(ofdef.NXM_NX_REG7, 1, 0, 1)
                                                                                         ]
                                                                            )
                                                                        ]
                                                                        ),
                                                                ofdef.ofp_instruction_goto_table(table_id = vi_next)
                                                                       ]
                                                       ))
                        cmds.append(ofdef.ofp_flow_mod(table_id = vo,
                                                       command = ofdef.OFPFC_ADD,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(
                                                                    oxm_fields = _create_oxms(
                                                                        _outport_oxm(pid)
                                                                        )
                                                                ),
                                                       instructions = [
                                                                ofdef.ofp_instruction_actions(
                                                                    actions = [
                                                                        ofdef.nx_action_resubmit(in_port = ofdef.nx_port_no.OFPP_IN_PORT,
                                                                                                 table = learning)
                                                                        ]
                                                                ),
                                                                ofdef.ofp_instruction_goto_table(table_id = vo_next)
                                                                ]))
                    elif not self._parent.prepush:
                        cmds.append(ofdef.ofp_flow_mod(table_id = vo,
                                                       cookie = 0x2,
                                                       cookie_mask = 0xffffffffffffffff,
                                                       command = ofdef.OFPFC_ADD,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(
                                                                    oxm_fields = _create_oxms(
                                                                        _outport_oxm(pid)
                                                                        )
                                                                ),
                                                       instructions = [
                                                                ofdef.ofp_instruction_actions(
                                                                    actions = [
                                                                        ofdef.ofp_action_output(port = ofdef.OFPP_CONTROLLER,
                                                                                                max_len = 32)
                                                                        ]
                                                                )]))                        
                    cmds.append(ofdef.ofp_flow_mod(table_id = egress,
                                                   command = ofdef.OFPFC_ADD,
                                                   priority = ofdef.OFP_DEFAULT_PRIORITY + 10,
                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                   out_port = ofdef.OFPP_ANY,
                                                   out_group = ofdef.OFPG_ANY,
                                                   match = ofdef.ofp_match_oxm(
                                                                oxm_fields = _create_oxms(
                                                                              _outport_oxm(pid),
                                                                              _flags(True, False)
                                                                            )
                                                            ),
                                                   instructions = [
                                                            ofdef.ofp_instruction_actions(
                                                                actions = [ofdef.ofp_action_output(port = pid)]
                                                            )
                                                        ]
                                                   ))
                    cmds.append(ofdef.ofp_flow_mod(table_id = egress,
                                                   command = ofdef.OFPFC_ADD,
                                                   priority = ofdef.OFP_DEFAULT_PRIORITY + 10,
                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                   out_port = ofdef.OFPP_ANY,
                                                   out_group = ofdef.OFPG_ANY,
                                                   match = ofdef.ofp_match_oxm(
                                                                oxm_fields = _create_oxms(
                                                                              _outport_oxm(pid),
                                                                              _flags(True, True)
                                                                            ) + [ofdef.create_oxm(ofdef.OXM_OF_IN_PORT, pid)]
                                                            ),
                                                   instructions = [
                                                            ofdef.ofp_instruction_actions(
                                                                actions = [ofdef.ofp_action_output(port = ofdef.OFPP_IN_PORT)]
                                                            )
                                                        ]
                                                   ))
            for m in self.execute_commands(conn, cmds):
                yield m
        subroutines.append(flow_mod())
        if self._parent.prepush or (not self._parent.prepush and not self._parent.learning):
            # prepush can be used together with learning
            if self._parent.prepush:
                # Never timeout
                idle_timeout = 0
                flags = 0
            else:
                idle_timeout = self._parent.pushtimeout
                flags = ofdef.OFPFF_SEND_FLOW_REM
            def flow_mod_prepush():
                remove_cmds = []
                add_cmds = []
                current_time = time()
                def _delete_flow(pid, nid, mac_address, network):
                    flows = [ofdef.ofp_flow_mod(table_id = vo,
                                               cookie = 0x3,
                                               cookie_mask = 0xffffffffffffffff,
                                               command = ofdef.OFPFC_DELETE_STRICT,
                                               priority = ofdef.OFP_DEFAULT_PRIORITY + 9,
                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                               out_port = ofdef.OFPP_ANY,
                                               out_group = ofdef.OFPG_ANY,
                                               match = ofdef.ofp_match_oxm(
                                                            oxm_fields = [ofdef.create_oxm(ofdef.NXM_NX_REG6, pid),
                                                                          ofdef.create_oxm(ofdef.NXM_NX_REG5, nid),
                                                                          ofdef.create_oxm(ofdef.OXM_OF_ETH_DST, ofdef.mac_addr(mac_address)),
                                                                          ]
                                                        )
                                               )]
                    if (network, mac_address) in self._buffered_packets:
                        # Drop any packets
                        packets = self._buffered_packets.pop((network, mac_address))
                        flows.extend(ofdef.ofp_packet_out(buffer_id = p[0],
                                                          in_port = p[3],
                                                          actions = []
                                                          )
                                     for p in packets
                                     if p[2] > current_time and p[0] != ofdef.OFP_NO_BUFFER)
                    return flows
                def _create_flow(pid, nid, mac_address, tunnelid, tunnel_dst, network, modify = False):
                    flows = [ofdef.ofp_flow_mod(table_id = vo,
                                                cookie = 0x3,
                                                cookie_mask = 0xffffffffffffffff,
                                                command = ofdef.OFPFC_MODIFY_STRICT if modify else ofdef.OFPFC_ADD, 
                                                priority = ofdef.OFP_DEFAULT_PRIORITY + 9,
                                                idle_timeout = idle_timeout,
                                                buffer_id = ofdef.OFP_NO_BUFFER,
                                                out_port = ofdef.OFPP_ANY,
                                                out_group = ofdef.OFPG_ANY,
                                                flags = flags,
                                                match = ofdef.ofp_match_oxm(
                                                             oxm_fields = [ofdef.create_oxm(ofdef.NXM_NX_REG6, pid),
                                                                           ofdef.create_oxm(ofdef.NXM_NX_REG5, nid),
                                                                           ofdef.create_oxm(ofdef.OXM_OF_ETH_DST, ofdef.mac_addr(mac_address)),
                                                                           ]
                                                         ),
                                                instructions = [
                                                        ofdef.ofp_instruction_actions(
                                                            actions = [ofdef.ofp_action_set_field(field = ofdef.create_oxm(ofdef.OXM_OF_TUNNEL_ID, tunnelid)),
                                                                       ofdef.ofp_action_set_field(field = ofdef.create_oxm(ofdef.NXM_NX_TUN_IPV4_DST, ofdef.ip4_addr(tunnel_dst))),
                                                                       ofdef.nx_action_reg_load(dst = ofdef.NXM_NX_REG7, ofs_nbits = ofdef.create_ofs_nbits(0, 1),
                                                                                                value = 1)]
                                                        ),
                                                        ofdef.ofp_instruction_goto_table(table_id = vo_next)
                                                    ]
                                               )]
                    if (network, mac_address) in self._watched_macs:
                        # Expire with idle time
                        self._watched_macs[(network, mac_address)] = None
                    if (network, mac_address) in self._buffered_packets:
                        # Send buffered packets
                        packets = self._buffered_packets.pop((network, mac_address))
                        flows.extend(ofdef.ofp_packet_out(buffer_id = p[0],
                                                          in_port = p[3],
                                                          actions = [ofdef.ofp_action_set_field(field = meta) for meta in p[4]]
                                                                    + [ofdef.ofp_action_set_field(field = ofdef.create_oxm(ofdef.OXM_OF_TUNNEL_ID, tunnelid)),
                                                                       ofdef.ofp_action_set_field(field = ofdef.create_oxm(ofdef.NXM_NX_TUN_IPV4_DST, ofdef.ip4_addr(tunnel_dst))),
                                                                       ofdef.nx_action_reg_load(dst = ofdef.NXM_NX_REG7, ofs_nbits = ofdef.create_ofs_nbits(0, 1),
                                                                                                value = 1),
                                                                       ofdef.nx_action_resubmit(in_port = ofdef.nx_port_no.OFPP_IN_PORT,
                                                                                              table = vo_next)],
                                                          data = p[1]
                                                          )
                                     for p in packets
                                     if p[2] > current_time and p[0] != ofdef.OFP_NO_BUFFER)
                    return flows
                for vxlaninfo, value in lastvxlaninfo.items():
                    _, lognet, nid, vni, mac_address, endpoint, _, pid = value
                    if vxlaninfo not in currentvxlaninfo:
                        self._parent._logger.debug("Found removed VXLAN endpoint: %r (%r -> %s), nid = %r, pid = %r, lognet = %r (vni = %r)",
                                                   vxlaninfo.id,
                                                   mac_address,
                                                   endpoint['tunnel_dst'],
                                                   nid,
                                                   pid,
                                                   lognet.id,
                                                   vni)
                        remove_cmds.extend(_delete_flow(pid, nid, mac_address, lognet))
                    else:
                        _, lognet2, nid2, vni2, mac_address2, endpoint2, _, pid2 = currentvxlaninfo[vxlaninfo]
                        if (pid2, nid2, mac_address2) != (pid, nid, mac_address):
                            self._parent._logger.debug("Found removed VXLAN endpoint: %r (%r -> %s), nid = %r, pid = %r, lognet = %r (vni = %r)",
                                                       vxlaninfo.id,
                                                       mac_address,
                                                       endpoint['tunnel_dst'],
                                                       nid,
                                                       pid,
                                                       lognet.id,
                                                       vni)
                            remove_cmds.extend(_delete_flow(pid, nid, mac_address, lognet))
                            self._parent._logger.debug("Found new VXLAN endpoint: %r (%r -> %s), nid = %r, pid = %r, lognet = %r (vni = %r)",
                                                       vxlaninfo.id,
                                                       mac_address2,
                                                       endpoint2['tunnel_dst'],
                                                       nid2,
                                                       pid2,
                                                       lognet2.id,
                                                       vni2)
                            add_cmds.extend(_create_flow(pid2, nid2, mac_address2, vni2, endpoint2['tunnel_dst'], lognet2))
                        else:
                            if vni == vni2 and endpoint['tunnel_dst'] == endpoint2['tunnel_dst']:
                                continue
                            self._parent._logger.debug("Found modified VXLAN endpoint: %r (%r -> %s), nid = %r, pid = %r, lognet = %r (vni = %r)",
                                                       vxlaninfo.id,
                                                       mac_address2,
                                                       endpoint2['tunnel_dst'],
                                                       nid2,
                                                       pid2,
                                                       lognet2.id,
                                                       vni2)
                            add_cmds.extend(_create_flow(pid2, nid2, mac_address2, vni2, endpoint2['tunnel_dst'], lognet2, True))
                for vxlaninfo, value in currentvxlaninfo.items():
                    if vxlaninfo not in lastvxlaninfo:
                        _, lognet, nid, vni, mac_address, endpoint, _, pid = value
                        self._parent._logger.debug("Found new VXLAN endpoint: %r (%r -> %s), nid = %r, pid = %r, lognet = %r (vni = %r)",
                                                   vxlaninfo.id,
                                                   mac_address,
                                                   endpoint['tunnel_dst'],
                                                   nid,
                                                   pid,
                                                   lognet.id,
                                                   vni)
                        add_cmds.extend(_create_flow(pid, nid, mac_address, vni, endpoint['tunnel_dst'], lognet))
                self._parent._logger.debug("Send %d remove commands and %d add commands", len(remove_cmds), len(add_cmds))
                for m in self.execute_commands(conn, remove_cmds):
                    yield m
                for m in self.execute_commands(conn, add_cmds):
                    yield m
            subroutines.append(flow_mod_prepush())            
        try:
            for m in self.executeAll(subroutines, retnames = ()):
                yield m
        except Exception:
            self._parent._logger.warning("Update vxlancast flow for connection %r failed with exception", conn, exc_info = True)

@defaultconfig
@depend(ofpportmanager.OpenflowPortManager, objectdb.ObjectDB, ovsdbportmanager.OVSDBPortManager)
class VXLANCast(FlowBase):
    "VXLAN single-cast and broadcast functions"
    _tablerequest = (("vxlaninput", ('l2input',), ''),
                     ("l3input", ("vxlaninput",), ''),
                     ("l2output", ("l3input",), ''),
                     ("vxlanoutput", ('l2output','vxlaninput'), ''),
                     ('egress', ('vxlanoutput', 'vxlanlearning'), ''),
                     ("vxlanlearning", ('vxlanoutput',), 'vxlanlearning'))
    # Enable learning on VXLAN tunnel IP destination.
    # VXLAN model must know every destination for each unicast MAC address. There are three
    # working mode the VXLANCast module:
    # 
    # 1. prepush = True, necessary MAC-tunneldst information will be pushed to the logical switch.
    #    learning = True may also be enabled so that external VXLAN MAC addresses may use learning
    #    strategy. If learning = False, unknown MAC addresses cannot be forwarded.
    # 
    # 2. learning = True, prepush = False. When a packet is received from a destination, its
    #    destination tunnel IP address is recorded, a flow is created for this MAC address;
    #    every packet with an unknown unicast MAC address as its destination will be broadcasted.
    # 
    # 3. learning = False, prepush = False. This enables a first-packet-to-controller strategy:
    #    When the first packet is going out from the switch, it is transmitted to the controller,
    #    and the controller queries ObjectDB for the tunnel destination, creates the fast-path
    #    and retransmit the packet. This may introduce a significant delay for the first packet.
    _default_learning = True
    # Enable prepush on VXLAN tunnel IP destination, the necessary MAC-tunneldst information
    # will be written into switch
    _default_prepush = True
    # Learning flow timeout
    _default_learntimeout = 300
    # When using first-packet-to-controller strategy, the buffered packets will only be keep for a
    # short time before expiring
    _default_packetexpire = 3
    # When using first-packet-to-controller strategy, the created flow will be timeout after this time
    _default_pushtimeout = 300
    # Mapping OpenFlow vHosts to OVSDB vHosts
    _default_vhostmap = {}
    # Every logical network has a broadcast list which contains all related server nodes,
    # this is the refresh interval to keep current node in the list
    _default_refreshinterval = 3600
    # When a logical port is migrating from one node to another, there may be multiple
    # instances of the logical port on different or same nodes. This is the maximum time
    # allowed for the migration.
    _default_allowedmigrationtime = 120
    def __init__(self, server):
        FlowBase.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._main
        self.routines.append(self.apiroutine)
        self._flowupdaters = {}
        self.createAPI(publicapi(self.createioflowparts, self.apiroutine,
                                 lambda connection,logicalnetwork,**kwargs:
                                        _is_vxlan(logicalnetwork)))
    def _main(self):
        flow_init = FlowInitialize.createMatcher(_ismatch = lambda x: self.vhostbind is None or x.vhost in self.vhostbind)
        conn_down = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_DOWN,
                                                               _ismatch = lambda x: self.vhostbind is None or x.createby.vhost in self.vhostbind)
        while True:
            yield (flow_init, conn_down)
            if self.apiroutine.matcher is flow_init:
                c = self.apiroutine.event.connection
                self.apiroutine.subroutine(self._init_conn(self.apiroutine.event.connection))
            else:
                c = self.apiroutine.event.connection
                self.apiroutine.subroutine(self._remove_conn(c))
    def _init_conn(self, conn):
        # Default
        ofdef = conn.openflowdef
        vhost = conn.protocol.vhost
        vi = self._gettableindex('vxlaninput', vhost)
        vi_next = self._getnexttable('', 'vxlaninput', vhost = vhost)
        vo = self._gettableindex('vxlanoutput', vhost)
        vo_next = self._getnexttable('', 'vxlanoutput', vhost = vhost)
        l3 = self._gettableindex('l3input', vhost)
        l3_next = self._getnexttable('', 'l3input', vhost = vhost)
        learning = self._gettableindex('vxlanlearning', vhost)
        if hasattr(conn, '_vxlancast_learning_routine') and conn._vxlancast_learning_routine:
            conn._vxlancast_learning_routine.close()
            delattr(conn, '_vxlancast_learning_routine')
        if conn in self._flowupdaters:
            vxlanupdater = self._flowupdaters.pop(conn)
            vxlanupdater.close()
        vxlanupdater = VXLANUpdater(conn, self)
        #flowupdater = VXLANFlowUpdater(conn, self)
        self._flowupdaters[conn] = vxlanupdater
        vxlanupdater.start()
        #flowupdater.start()
        cmds = [ofdef.ofp_flow_mod(table_id = vo,
                                    command = ofdef.OFPFC_ADD,
                                    priority = ofdef.OFP_DEFAULT_PRIORITY + 10,
                                    buffer_id = ofdef.OFP_NO_BUFFER,
                                    out_port = ofdef.OFPP_ANY,
                                    out_group = ofdef.OFPG_ANY,
                                    match = ofdef.ofp_match_oxm(
                                                oxm_fields = [
                                                    # A broadcast packet
                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST_W, b'\x01\x00\x00\x00\x00\x00', b'\x01\x00\x00\x00\x00\x00')
                                                    ]
                                            ),
                                    instructions = [ofdef.ofp_instruction_goto_table(table_id = vo_next)]
                                    )]
        if not conn.protocol.disablenxext and self.learning:
            cmds.append(ofdef.ofp_flow_mod(table_id = learning,
                                           out_port = ofdef.OFPP_ANY,
                                           out_group = ofdef.OFPG_ANY,
                                           command = ofdef.OFPFC_ADD,
                                           priority = 0,
                                           buffer_id = ofdef.OFP_NO_BUFFER,
                                           match = ofdef.ofp_match_oxm(),
                                           instructions = [ofdef.ofp_instruction_actions(
                                                                actions = [ofdef.nx_action_reg_load(dst = ofdef.NXM_NX_REG7,
                                                                                                ofs_nbits = ofdef.create_ofs_nbits(0, 1),
                                                                                                value = 0)]
                                                            )]))
        for m in conn.protocol.batch(cmds, conn, self.apiroutine):
            yield m
    def _remove_conn(self, conn):
        # Do not need to modify flows
        if conn in self._flowupdaters:
            vxlanupdater = self._flowupdaters.pop(conn)
            vxlanupdater.close()
        if False:
            yield
    def createioflowparts(self,connection,logicalnetwork,physicalport,logicalnetworkid,physicalportid):
        #
        #  1. used in IOProcessing , when physicalport add to logicalnetwork 
        #     return : input flow match vxlan vni, input flow vlan parts actions
        #              output flow vxlan parts actions, output group bucket
        #
        group_created = False
        group_portid = None
        ofdef = connection.openflowdef
        if connection in self._flowupdaters:
            # Wait for group creation
            vxlanupdater = self._flowupdaters[connection]
            try:
                for m in vxlanupdater.wait_for_group(self.apiroutine, logicalnetworkid):
                    yield m
            except Exception:
                self._logger.warning("Group is not created, connection = %r, logicalnetwork = %r, logicalnetworkid = %r", connection, logicalnetwork, logicalnetworkid, exc_info = True)
            else:
                group_created = True
                group_portid = self.apiroutine.retvalue
        if not group_created:
            self.apiroutine.retvalue = ([ofdef.create_oxm(ofdef.OXM_OF_TUNNEL_ID, getattr(logicalnetwork, 'vni', 0))],
                                        [],
                                        [],
                                        [],
                                        [])
        # We are removing this because the physical port may change. If there are more than one physical port
        # in the physical network, the packets may be sent more than once,
        # but the solution is simple: DON'T DO THAT
        # elif group_portid == physicalportid:
        else:
            self.apiroutine.retvalue = ([ofdef.create_oxm(ofdef.OXM_OF_TUNNEL_ID, getattr(logicalnetwork, 'vni', 0))],
                                        [],
                                        [ofdef.ofp_action_set_field(field = ofdef.create_oxm(ofdef.OXM_OF_TUNNEL_ID, getattr(logicalnetwork, 'vni', 0))),
                                         ofdef.ofp_action_group(group_id = (logicalnetworkid & 0xffff) | 0x10000)],
                                        [ofdef.ofp_action_set_field(field = ofdef.create_oxm(ofdef.OXM_OF_TUNNEL_ID, getattr(logicalnetwork, 'vni', 0))),
                                         ofdef.ofp_action_group(group_id = (logicalnetworkid & 0xffff) | 0x10000)],
                                        [ofdef.ofp_action_set_field(field = ofdef.create_oxm(ofdef.OXM_OF_TUNNEL_ID, getattr(logicalnetwork, 'vni', 0))),
                                         ofdef.ofp_action_group(group_id = (logicalnetworkid & 0xffff) | 0x20000)],
                                        )
        #=======================================================================
        # else:
        #     self.apiroutine.retvalue = ([ofdef.create_oxm(ofdef.OXM_OF_TUNNEL_ID, getattr(logicalnetwork, 'vni', 0))],
        #                                 [],
        #                                 [ofdef.ofp_action_set_field(field = ofdef.create_oxm(ofdef.OXM_OF_TUNNEL_ID, getattr(logicalnetwork, 'vni', 0))),
        #                                  ofdef.ofp_action_group(group_id = (logicalnetworkid & 0xffff) | 0x10000)],
        #                                 [],
        #                                 [ofdef.ofp_action_set_field(field = ofdef.create_oxm(ofdef.OXM_OF_TUNNEL_ID, getattr(logicalnetwork, 'vni', 0))),
        #                                 ofdef.ofp_action_group(group_id = (logicalnetworkid & 0xffff) | 0x10000)]
        #                                 )
        #=======================================================================




'''
Created on 2016/8/25

:author: hubo
'''

from namedstruct import *
from namedstruct.namedstruct import BadFormatError, BadLenError, Parser, _create_struct

def _tobytes(s, encoding = 'utf-8'):
    if s is None:
        return None
    elif isinstance(s, bytes):
        return s
    else:
        return s.encode(encoding)

class UStringParser(object):
    '''
    Jute ustring type. 
    '''
    def __init__(self):
        pass
    def parse(self, buffer, inlineparent = None):
        if len(buffer) < 4:
            return None
        length = int32.create(buffer[:4])
        if length < 0:
            return (None, 4)
        if len(buffer) < 4 + length:
            return None
        else:
            return (buffer[4:4+length], 4 + length)
    def new(self, inlineparent = None):
        return b''
    def create(self, data, inlineparent = None):
        r = self.parse(data)
        if r is None:
            raise BadLenError('Ustring length not match')
        return r[0]
    def sizeof(self, prim):
        prim = _tobytes(prim)
        if prim is None:
            return 4
        else:
            return len(prim) + 4
    def paddingsize(self, prim):
        return self.sizeof(prim)
    def tobytes(self, prim, skipprepack = False):
        prim = _tobytes(prim)
        if prim is None:
            return int32.tobytes(-1)
        else:
            return int32.tobytes(len(prim)) + prim

class ustringtype(typedef):
    '''
    A int32 followed by variable length bytes
    '''
    _parser = UStringParser()
    def __init__(self, displayname = 'ustring'):
        typedef.__init__(self)
        self._displayname = displayname
    def parser(self):
        return self._parser
    def __repr__(self, *args, **kwargs):
        return self._displayname

ustring = ustringtype()
z_buffer = ustringtype('buffer')


class VectorParser(object):
    '''
    Jute vector type. 
    '''
    def __init__(self, innerparser):
        self._innerparser = innerparser
    def parse(self, buffer, inlineparent = None):
        if len(buffer) < 4:
            return None
        length = int32.create(buffer[:4])
        if length < 0:
            return (None, 4)
        start = 4
        result = []
        for i in range(0, length):
            r = self._innerparser.parse(buffer[start:], None)
            if r is None:
                return None
            (inner, size) = r
            result.append(inner)
            start += size
        return (result, start)
    def new(self, inlineparent = None):
        return []
    def create(self, data, inlineparent = None):
        r = self.parse(data)
        if r is None:
            raise BadLenError('Ustring length not match')
        return r[0]
    def sizeof(self, prim):
        if prim is None:
            return 4
        else:
            return sum(self._innerparser.paddingsize(r) for r in prim) + 4
    def paddingsize(self, prim):
        return self.sizeof(prim)
    def tobytes(self, prim, skipprepack = False):
        if prim is None:
            return int32.tobytes(-1)
        else:
            return int32.tobytes(len(prim)) + b''.join(self._innerparser.tobytes(r) for r in prim)

class vector(typedef):
    '''
    Jute vector
    '''
    def __init__(self, innertype):
        typedef.__init__(self)
        self._innertype = innertype
    def _compile(self):
        return VectorParser(self._innertype.parser())
    def __repr__(self, *args, **kwargs):
        return 'vector<' + repr(self._innertype) + '>'

# /* predefined xid's values recognized as special by the server */
zk_xid = enum('zk_xid', globals(), int32,
    WATCHER_EVENT_XID = -1, 
    PING_XID = -2,
    AUTH_XID = -4,
    SET_WATCHES_XID = -8)

# /* zookeeper event type constants */
zk_watch_event = enum('zk_watch_event', globals(), int32,
    CREATED_EVENT_DEF = 1,
    DELETED_EVENT_DEF = 2,
    CHANGED_EVENT_DEF = 3,
    CHILD_EVENT_DEF = 4,
    SESSION_EVENT_DEF = -1,
    NOTWATCHING_EVENT_DEF = -2)

zk_request_type = enum('zk_request_type', globals(), int32,
    ZOO_NOTIFY_OP = 0,
    ZOO_CREATE_OP = 1,
    ZOO_DELETE_OP = 2,
    ZOO_EXISTS_OP = 3,
    ZOO_GETDATA_OP = 4,
    ZOO_SETDATA_OP = 5,
    ZOO_GETACL_OP = 6,
    ZOO_SETACL_OP = 7,
    ZOO_GETCHILDREN_OP = 8,
    ZOO_SYNC_OP = 9,
    ZOO_PING_OP = 11,
    ZOO_GETCHILDREN2_OP = 12,
    ZOO_CHECK_OP = 13,
    ZOO_MULTI_OP = 14,
    ZOO_CREATE_SESSION_OP = -10,
    ZOO_CLOSE_SESSION_OP = -11,
    ZOO_SETAUTH_OP = 100,
    ZOO_SETWATCHES_OP = 101,
    ZOO_SASL_OP = 102,           # NOT SUPPORTED
    ZOO_ERROR_TYPE = -1
)

zk_client_state = enum('zk_client_state', globals(), int32,
    ZOO_DISCONNECTED_STATE = 0,
    ZOO_NOSYNC_CONNECTED_STATE = 1,
    ZOO_SYNC_CONNECTED_STATE = 3,
    ZOO_AUTH_FAILED_STATE = 4,
    ZOO_CONNECTED_READONLY_STATE = 5,
    ZOO_SASL_AUTHENTICATED_STATE = 6,
    ZOO_EXPIRED_STATE = -112
)

zk_err = enum('zk_err', globals(), int32,
    ZOO_ERR_OK = 0,
    ZOO_ERR_SYSTEMERROR = -1,
    ZOO_ERR_RUNTIMEINCONSISTENCY = -2,
    ZOO_ERR_DATAINCONSISTENCY = -3,
    ZOO_ERR_CONNECTIONLOSS = -4,
    ZOO_ERR_MARSHALLINGERROR = -5,
    ZOO_ERR_UNIMPLEMENTED = -6,
    ZOO_ERR_OPERATIONTIMEOUT = -7,
    ZOO_ERR_BADARGUMENTS = -8,
    ZOO_ERR_APIERROR = -100,
    ZOO_ERR_NONODE = -101,
    ZOO_ERR_NOAUTH = -102,
    ZOO_ERR_BADVERSION = -103,
    ZOO_ERR_NOCHILDRENFOREPHEMERALS = -108,
    ZOO_ERR_NODEEXISTS = -110,
    ZOO_ERR_NOTEMPTY = -111,
    ZOO_ERR_SESSIONEXPIRED = -112,
    ZOO_ERR_INVALIDCALLBACK = -113,
    ZOO_ERR_INVALIDACL = -114,
    ZOO_ERR_AUTHFAILED = -115
    )

zk_perm = enum('zk_perm', globals(), int32, True,
    ZOO_PERM_READ = 1 << 0,
    ZOO_PERM_WRITE = 1 << 1,
    ZOO_PERM_CREATE = 1 << 2,
    ZOO_PERM_DELETE = 1 << 3,
    ZOO_PERM_ADMIN = 1 << 4,
    ZOO_PERM_ALL = 0x1f
)

zk_create_flag = enum('zk_create_flag', globals(), int32, True,
                    ZOO_EPHEMERAL = 1 << 0,
                    ZOO_SEQUENCE = 1 << 1,
    )

Id = nstruct(
        (ustring, 'scheme'),
        (ustring, 'id'),
        name = 'Id',
        padding = 1
     )

ACL = nstruct(
        (zk_perm, 'perms'),
        (Id, 'id'),
        name = 'ACL',
        padding = 1
    )
Stat = nstruct(
        (int64, 'czxid'),      # created zxid
        (int64, 'mzxid'),      # last modified zxid
        (int64, 'ctime'),      # created
        (int64, 'mtime'),      # last modified
        (int32, 'version'),     # version
        (int32, 'cversion'),    # child version
        (int32, 'aversion'),    # acl version
        (int64, 'ephemeralOwner'), # owner id if ephemeral, 0 otw
        (int32, 'dataLength'),  #length of the data in the node
        (int32, 'numChildren'), #number of children of this node
        (int64, 'pzxid'),      # last modified children
        name = 'Stat',
        padding = 1
    )
# information explicitly stored by the server persistently
StatPersisted = nstruct(
        (int64, 'czxid'),      # created zxid
        (int64, 'mzxid'),      # last modified zxid
        (int64, 'ctime'),      # created
        (int64, 'mtime'),      # last modified
        (int32, 'version'),     # version
        (int32, 'cversion'),    # child version
        (int32, 'aversion'),    # acl version
        (int64, 'ephemeralOwner'), # owner id if ephemeral, 0 otw
        (int64, 'pzxid'),      # last modified children
        name = 'StatPersisted',
        padding = 1
    )

# information explicitly stored by the version 1 database of servers 
StatPersistedV1 = nstruct(
       (int64, 'czxid'), #created zxid
       (int64, 'mzxid'), #last modified zxid
       (int64, 'ctime'), #created
       (int64, 'mtime'), #last modified
       (int32, 'version'), #version
       (int32, 'cversion'), #child version
       (int32, 'aversion'), #acl version
       (int64, 'ephemeralOwner'), #owner id if ephemeral. 0 otw
       name = 'StatPersistedV1',
       padding = 1
    )

def _pack_zookeeper_length(x):
    x.length = len(x) - 4

# Every message begin with 32-bit length, excluding the field it self
ZooKeeperRequest = nstruct(
            (uint32, 'length'),
            name = 'ZooKeeperRequest',
            padding = 1,
            size = lambda x: x.length + 4,
            prepack = _pack_zookeeper_length
        )

# ZooKeeper types parsing depends on the parser state, set the "hidden field" and use _autosubclass

_TypedZooKeeperRequest = nstruct(
            name = '_TypedZooKeeperRequest',
            base = ZooKeeperRequest,
            criteria = lambda x: hasattr(x, 'zookeeper_type')
            )

ZooKeeperReply = nstruct(
            (uint32, 'length'),
            name = 'ZooKeeperRequest',
            padding = 1,
            size = lambda x: x.length + 4,
            prepack = _pack_zookeeper_length
        )

_TypedZooKeeperReply = nstruct(
            name = '_TypedZooKeeperReply',
            base = ZooKeeperReply,
            criteria = lambda x: hasattr(x, 'zookeeper_type')
        )

CONNECT_PACKET = 0x1

ConnectRequest = nstruct(
        (int32, 'protocolVersion'),
        (int64, 'lastZxidSeen'),
        (int32, 'timeOut'),
        (int64, 'sessionId'),
        (z_buffer, 'passwd'),
        (boolean, 'readOnly'),
        name = 'ConnectRequest',
        base = _TypedZooKeeperRequest,
        criteria = lambda x: x.zookeeper_type == CONNECT_PACKET,
        init = packvalue(CONNECT_PACKET, 'zookeeper_type')
    )

_ConnectResponseOptional = nstruct(
        name = '_ConnectResponseOptional',
        padding = 1,
        inline = False
    )

_ConnectResponseReadOnly = nstruct(
        (boolean, 'readOnly'),
        name = '_ConnectResponseReadOnly',
        base = _ConnectResponseOptional,
        criteria = lambda x: x._realsize() > 0
    )

ConnectResponse = nstruct(
        (int32, 'protocolVersion'),
        (int32, 'timeOut'),
        (int64, 'sessionId'),
        (z_buffer, 'passwd'),
        (_ConnectResponseOptional,),
        name = 'ConnectResponse',
        base = _TypedZooKeeperReply,
        criteria = lambda x: x.zookeeper_type == CONNECT_PACKET,
        init = packvalue(CONNECT_PACKET, 'zookeeper_type'),
        lastextra = True
    )


# All the other requests and responses are prepended by a header

HEADER_PACKET = 0x2

RequestHeader = nstruct(
        (zk_xid, 'xid'),
        (zk_request_type, 'type'),
        name = 'RequestHeader',
        base = _TypedZooKeeperRequest,
        criteria = lambda x: x.zookeeper_type != CONNECT_PACKET,
        init = packvalue(HEADER_PACKET, 'zookeeper_type'),
        classifier = lambda x: x.type
    )

ReplyHeader = nstruct(
        (int32, 'xid'),
        (int64, 'zxid'),
        (zk_err, 'err'),
        name = 'ReplyHeader',
        base = _TypedZooKeeperReply,
        criteria = lambda x: x.zookeeper_type != CONNECT_PACKET,
        init = packvalue(HEADER_PACKET, 'zookeeper_type')
    )

# ZooKeeper reply type must be determined by the request type

_TypedReply = nstruct(
        name = '_TypedReply',
        base = ReplyHeader,
        criteria = lambda x: hasattr(x, 'zookeeper_request_type') and (x.err == 0 or len(x) > 20),
        classifier = lambda x: x.zookeeper_request_type
    )

SetWatches = nstruct(
        (int64, 'relativeZxid'),
        (vector(ustring), 'dataWatches'),
        (vector(ustring), 'existWatches'),
        (vector(ustring), 'childWatches'),
        name = 'SetWatches',
        base = RequestHeader,
        classifyby = (ZOO_SETWATCHES_OP,),
        init = lambda x: (packvalue(ZOO_SETWATCHES_OP, 'type')(x),
                          packvalue(SET_WATCHES_XID, 'xid')(x))
    )

AuthPacket = nstruct(
        (int32, 'auth_type'),             # This is not used, always 0
        (ustring, 'scheme'),
        (z_buffer, 'auth'),
        name = 'AuthPacket',
        base = RequestHeader,
        classifyby = (ZOO_SETAUTH_OP,),
        init = lambda x: (packvalue(ZOO_SETAUTH_OP, 'type')(x),
                          packvalue(AUTH_XID, 'xid')(x))
    )

MultiHeader = nstruct(
        (zk_request_type, 'type'),
        (boolean, 'done'),
        (zk_err, 'err'),
        name = 'MultiHeader',
        padding = 1
    )

_GetDataRequest = nstruct(
        (ustring, 'path'),
        (boolean, 'watch'),
        name = '_GetDataRequest',
        padding = 1
    )

_SetDataRequest = nstruct(
        (ustring, 'path'),
        (z_buffer, 'data'),
        (int32, 'version'),
        name = '_SetDataRequest',
        padding = 1
    )

_SetDataResponse = nstruct(
        (Stat, 'stat'),
        name = '_SetDataResponse',
        padding = 1
    )

_GetSASLRequest = nstruct(
        (z_buffer, 'token'),
        name = '_GetSASLRequest',
        padding = 1
    )

_SetSASLRequest = nstruct(
        (z_buffer, 'token'),
        name = '_SetSASLRequest',
        padding = 1
    )

_SetSASLResponse = nstruct(
        (z_buffer, 'token'),
        name = '_SetSASLResponse',
        padding = 1
    )

_CreateRequest = nstruct(
        (ustring, 'path'),
        (z_buffer, 'data'),
        (vector(ACL), 'acl'),
        (zk_create_flag, 'flags'),
        name = '_CreateRequest',
        padding = 1
    )

_DeleteRequest = nstruct(
        (ustring, 'path'),
        (int32, 'version'),
        name = '_DeleteRequest',
        padding = 1
    )

_GetChildrenRequest = nstruct(
        (ustring, 'path'),
        (boolean, 'watch'),
        name = '_GetChildrenRequest',
        padding = 1
    )

_GetChildren2Request = nstruct(
        (ustring, 'path'),
        (boolean, 'watch'),
        name = '_GetChildren2Request',
        padding = 1
    )

_CheckVersionRequest = nstruct(
        (ustring, 'path'),
        (int32, 'version'),
        name = '_CheckVersionRequest',
        padding = 1
    )

_GetMaxChildrenRequest = nstruct(
        (ustring, 'path'),
        name = '_GetMaxChildrenRequest',
        padding = 1
    )

_GetMaxChildrenResponse = nstruct(
        (int32, 'max'),
        name = '_GetMaxChildrenResponse',
        padding = 1
    )

_SetMaxChildrenRequest = nstruct(
        (ustring, 'path'),
        (int32, 'max'),
        name = '_SetMaxChildrenRequest',
        padding = 1
    )

_SyncRequest = nstruct(
        (ustring, 'path'),
        name = '_SyncRequest',
        padding = 1
    )

_SyncResponse = nstruct(
        (ustring, 'path'),
        name = '_SyncResponse',
        padding = 1
    )

_GetACLRequest = nstruct(
        (ustring, 'path'),
        name = '_GetACLRequest',
        padding = 1
    )

_SetACLRequest = nstruct(
        (ustring, 'path'),
        (vector(ACL), 'acl'),
        (int32, 'version'),
        name = '_SetACLRequest',
        padding = 1
    )

_SetACLResponse = nstruct(
        (Stat, 'stat'),
        name = '_SetACLResponse',
        padding = 1
    )

_WatcherEvent = nstruct(
        (zk_watch_event, 'type'),  # event type
        (zk_client_state, 'state'), # state of the Keeper client runtime
        (ustring, 'path'),
        name = '_WatcherEvent',
        padding = 1
    )

_ErrorResponse = nstruct(
        (zk_err, 'err2'),
        name = '_ErrorResponse',
        padding = 1
    )

_CreateResponse = nstruct(
        (ustring, 'path'),
        name = '_CreateResponse',
        padding = 1
    )

_ExistsRequest = nstruct(
        (ustring, 'path'),
        (boolean, 'watch'),
        name = '_ExistsRequest',
        padding = 1
    )

_ExistsResponse = nstruct(
        (Stat, 'stat'),
        name = '_ExistsResponse',
        padding = 1
    )

_GetDataResponse = nstruct(
        (z_buffer, 'data'),
        (Stat, 'stat'),
        name = '_GetDataResponse',
        padding = 1
    )

_GetChildrenResponse = nstruct(
        (vector(ustring), 'children'),
        name = '_GetChildrenResponse',
        padding = 1
    )

_GetChildren2Response = nstruct(
        (vector(ustring), 'children'),
        (Stat, 'stat'),
        name = '_GetChildren2Response',
        padding = 1
    )

_GetACLResponse = nstruct(
        (vector(ACL), 'acl'),
        (Stat, 'stat'),
        name = '_GetACLResponse',
        padding = 1
    )

# Requests

CreateRequest = nstruct(
        (_CreateRequest,),
        name = 'CreateRequest',
        base = RequestHeader,
        init = packvalue(ZOO_CREATE_OP, 'type'),
        classifyby = (ZOO_CREATE_OP,)
    )

CreateResponse = nstruct(
        (_CreateResponse,),
        name = 'CreateResponse',
        base = _TypedReply,
        init = packvalue(ZOO_CREATE_OP, 'zookeeper_request_type'),
        classifyby = (ZOO_CREATE_OP,)
    )

DeleteRequest = nstruct(
        (_DeleteRequest,),
        name = 'DeleteRequest',
        base = RequestHeader,
        init = packvalue(ZOO_DELETE_OP, 'type'),
        classifyby = (ZOO_DELETE_OP,)
    )

# DeleteReply = ReplyHeader

ExistsRequest = nstruct(
        (_ExistsRequest,),
        name = 'ExistsRequest',
        base = RequestHeader,
        init = packvalue(ZOO_EXISTS_OP, 'type'),
        classifyby = (ZOO_EXISTS_OP,)
    )

ExistsResponse = nstruct(
        (_ExistsResponse,),
        name = 'ExistsResponse',
        base = _TypedReply,
        init = packvalue(ZOO_EXISTS_OP, 'zookeeper_request_type'),
        classifyby = (ZOO_EXISTS_OP,)
    )

GetDataRequest = nstruct(
        (_GetDataRequest,),
        base = RequestHeader,
        name = 'GetDataRequest',
        init = packvalue(ZOO_GETDATA_OP, 'type'),
        classifyby = (ZOO_GETDATA_OP,)
    )

GetDataResponse = nstruct(
        (_GetDataResponse,),
        base = _TypedReply,
        name = 'GetDataResponse',
        init = packvalue(ZOO_GETDATA_OP, 'zookeeper_request_type'),
        classifyby = (ZOO_GETDATA_OP,)
    )

SetDataRequest = nstruct(
        (_SetDataRequest,),
        base = RequestHeader,
        name = 'SetDataRequest',
        init = packvalue(ZOO_SETDATA_OP, 'type'),
        classifyby = (ZOO_SETDATA_OP,)
    )

SetDataResponse = nstruct(
        (_SetDataResponse,),
        base = _TypedReply,
        name = 'SetDataResponse',
        init = packvalue(ZOO_SETDATA_OP, 'zookeeper_request_type'),
        classifyby = (ZOO_SETDATA_OP,)
    )

SetSASLRequest = nstruct(
        (_SetSASLRequest,),
        base = RequestHeader,
        name = 'SetSASLRequest',
        init = packvalue(ZOO_SASL_OP, 'type'),
        classifyby = (ZOO_SASL_OP,)
    )

GetSASLRequest = SetSASLRequest

SetSASLResponse = nstruct(
        (_SetSASLResponse,),
        base = _TypedReply,
        name = 'SetSASLResponse',
        init = packvalue(ZOO_SASL_OP, 'zookeeper_request_type'),
        classifyby = (ZOO_SASL_OP,)
    )

GetSASLResponse = SetSASLResponse

GetACLRequest = nstruct(
        (_GetACLRequest,),
        base = RequestHeader,
        name = 'GetACLRequest',
        init = packvalue(ZOO_GETACL_OP, 'type'),
        classifyby = (ZOO_GETACL_OP,)
    )

GetACLResponse = nstruct(
        (_GetACLResponse,),
        base = _TypedReply,
        name = 'GetACLResponse',
        init = packvalue(ZOO_GETACL_OP, 'zookeeper_request_type'),
        classifyby = (ZOO_GETACL_OP,)
    )

SetACLRequest = nstruct(
        (_SetACLRequest,),
        base = RequestHeader,
        name = 'SetACLRequest',
        init = packvalue(ZOO_SETACL_OP, 'type'),
        classifyby = (ZOO_SETACL_OP,)
    )

SetACLResponse = nstruct(
        (_SetACLResponse,),
        base = _TypedReply,
        name = 'SetACLResponse',
        init = packvalue(ZOO_SETACL_OP, 'zookeeper_request_type'),
        classifyby = (ZOO_SETACL_OP,)
    )

GetChildrenRequest = nstruct(
        (_GetChildrenRequest,),
        base = RequestHeader,
        name = 'GetChildrenRequest',
        init = packvalue(ZOO_GETCHILDREN_OP, 'type'),
        classifyby = (ZOO_GETCHILDREN_OP,)
    )

GetChildrenResponse = nstruct(
        (_GetChildrenResponse,),
        base = _TypedReply,
        name = 'GetChildrenResponse',
        init = packvalue(ZOO_GETCHILDREN_OP, 'zookeeper_request_type'),
        classifyby = (ZOO_GETCHILDREN_OP,)
    )

SyncRequest = nstruct(
        (_SyncRequest,),
        base = RequestHeader,
        name = 'SyncRequest',
        init = packvalue(ZOO_SYNC_OP, 'type'),
        classifyby = (ZOO_SYNC_OP,)
    )

SyncResponse = nstruct(
        (_SyncResponse,),
        base = _TypedReply,
        name = 'SyncResponse',
        init = packvalue(ZOO_SYNC_OP, 'zookeeper_request_type'),
        classifyby = (ZOO_SYNC_OP,)
    )

# Ping Request and Response have no body

GetChildren2Request = nstruct(
        (_GetChildren2Request,),
        base = RequestHeader,
        name = 'GetChildren2Request',
        init = packvalue(ZOO_GETCHILDREN2_OP, 'type'),
        classifyby = (ZOO_GETCHILDREN2_OP,)
    )

GetChildren2Response = nstruct(
        (_GetChildren2Response,),
        base = _TypedReply,
        name = 'GetChildren2Response',
        init = packvalue(ZOO_GETCHILDREN2_OP, 'zookeeper_request_type'),
        classifyby = (ZOO_GETCHILDREN2_OP,)
    )

MultiOpRequest = nvariant(
        'MultiOpRequest',
        MultiHeader,
        classifier = lambda x: x.type,
        padding = 1
    )

MultiOpResponse = nvariant(
        'MultiOpResponse',
        MultiHeader,
        classifier = lambda x: x.type,
        padding = 1
    )

MultiOpCheck = nstruct(
        (_CheckVersionRequest,),
        name = 'MultiOpCheck',
        base = MultiOpRequest,
        init = packvalue(ZOO_CHECK_OP, 'type'),
        classifyby = (ZOO_CHECK_OP,)
    )

# Check does not have response body

MultiOpCreate = nstruct(
        (_CreateRequest,),
        name = 'MultiOpCreate',
        base = MultiOpRequest,
        init = packvalue(ZOO_CREATE_OP, 'type'),
        classifyby = (ZOO_CREATE_OP,)
    )

MultiOpCreateResponse = nstruct(
        (_CreateResponse,),
        name = 'MultiOpCreateResponse',
        base = MultiOpResponse,
        init = packvalue(ZOO_CREATE_OP, 'type'),
        classifyby = (ZOO_CREATE_OP,)
    )

MultiOpDelete = nstruct(
        (_DeleteRequest,),
        name = 'MultiOpDelete',
        base = MultiOpRequest,
        init = packvalue(ZOO_DELETE_OP, 'type'),
        classifyby = (ZOO_DELETE_OP,)
    )

MultiOpSetData = nstruct(
        (_SetDataRequest,),
        name = 'MultiOpSetData',
        base = MultiOpRequest,
        init = packvalue(ZOO_SETDATA_OP, 'type'),
        classifyby = (ZOO_SETDATA_OP,)
    )

MultiOpSetDataResponse = nstruct(
        (_SetDataResponse,),
        name = 'MultiOpSetDataResponse',
        base = MultiOpResponse,
        init = packvalue(ZOO_SETDATA_OP, 'type'),
        classifyby = (ZOO_SETDATA_OP,)
    )

MultiOpErrorResponse = nstruct(
        (_ErrorResponse,),
        name = 'MultiOpErrorResponse',
        base = MultiOpResponse,
        init = packvalue(ZOO_ERROR_TYPE, 'type'),
        classifyby = (ZOO_ERROR_TYPE,)
    )

MultiRequest = nstruct(
        (MultiOpRequest[0], 'requests'),
        name = 'MultiRequest',
        base = RequestHeader,
        classifyby = (ZOO_MULTI_OP,),
        init = packvalue(ZOO_MULTI_OP, 'type')
    )



MultiResponse = nstruct(
        (MultiOpResponse[0], 'responses'),
        name = 'MultiResponse',
        base = _TypedReply,
        classifyby = (ZOO_MULTI_OP,),
        init = packvalue(ZOO_MULTI_OP, 'zookeeper_request_type')
    )

WatcherEvent = nstruct(
        (_WatcherEvent,),
        name = 'WatcherEvent',
        base = ReplyHeader,
        criteria = lambda x: x.xid == WATCHER_EVENT_XID,
        init = packvalue(WATCHER_EVENT_XID, 'xid')
    )

# Helpers

def default_acl():
    return ACL(perms = ZOO_PERM_ALL, id = Id(scheme = 'world', id = 'anyone'))

def multi(*ops):
    return MultiRequest(
                requests = list(ops) + \
                            [MultiOpRequest(type = -1, done = True, err = -1)]
                )

def create(path, data, ephemeral = False, sequence = False, acl = None):
    if acl is None:
        acl = [default_acl()]
    return CreateRequest(path = path, data = data, acl = acl,
                         flags = (ZOO_EPHEMERAL if ephemeral else 0) | (ZOO_SEQUENCE if sequence else 0))



def delete(path, version = -1):
    return DeleteRequest(path = path, version = -1)

def exists(path, watch = False):
    return ExistsRequest(path = path, watch = watch)

def getdata(path, watch = False):
    return GetDataRequest(path = path, watch = watch)

def setdata(path, data, version = -1):
    return SetDataRequest(path = path, data = data, version = -1)

def getchildren(path, watch = False):
    return GetChildrenRequest(path = path, watch = watch)

def getchildren2(path, watch = False):
    return GetChildren2Request(path = path, watch = watch)

def multi_create(path, data, ephemeral = False, sequence = False, acl = None):
    if acl is None:
        acl = [default_acl()]
    return MultiOpCreate(path = path, data = data, acl = acl,
                         flags = (ZOO_EPHEMERAL if ephemeral else 0) | (ZOO_SEQUENCE if sequence else 0))

def multi_delete(path, version = -1):
    return MultiOpDelete(path = path, version = -1)

def multi_check(path, version):
    return MultiOpCheck(path, version)

def multi_setdata(path, data, version = -1):
    return MultiOpSetData(path = path, data = data, version = -1)

def getacl(path):
    return GetACLRequest(path = path)

def setacl(path, acl, version = -1):
    return SetACLRequest(path = path, acl = acl, version = version)

def sync(path):
    return SyncRequest(path = path)




'''
Created on 2015/12/25

:author: hubo
'''

from vlcp.protocol import Protocol
from vlcp.config import defaultconfig
from vlcp.event import Event, withIndices, ConnectionWriteEvent
import logging
import os
from vlcp.event.stream import Stream, StreamDataEvent
from vlcp.event.connection import Client

@withIndices('state', 'connection', 'connmark', 'createby')
class RawConnectionStateEvent(Event):
    CONNECTION_UP = 'up'
    CONNECTION_DOWN = 'down'
    CONNECTION_NOTCONNECTED = 'notconnected'

def _copy(buffer):
    try:
        if isinstance(buffer, memoryview):
            return buffer.tobytes()
        else:
            return buffer[:]
    except:
        return buffer[:]

@defaultconfig
class Raw(Protocol):
    '''
    Raw protocol, provide two streams for input and output
    '''
    _default_persist = False
    _default_defaultport = 0
    _default_createqueue = True
    # Enable/disable buffering for the output stream.
    # It is dangerous to use buffering in output stream because small amount of data might
    # stay in buffer and not be sent
    _default_buffering = False
    _default_writebufferlimit = 4096
    # Split very large data to chunks to balance the output streaming
    _default_splitsize = 1048576
    _logger = logging.getLogger(__name__ + '.JsonRPC')
    def __init__(self):
        '''
        Constructor
        '''
        Protocol.__init__(self)
    def init(self, connection):
        for m in Protocol.init(self, connection):
            yield m
        connection.createdqueues.append(connection.scheduler.queue.addSubQueue(\
                self.messagepriority, RawConnectionStateEvent.createMatcher(connection = connection), ('connstate', connection)))
        for m in self.reconnect_init(connection):
            yield m
    def _raw_writer(self, connection):
        try:
            while True:
                for m in connection.outputstream.prepareRead(connection):
                    yield m
                try:
                    data = connection.outputstream.readonce()
                except EOFError:
                    for m in connection.write(ConnectionWriteEvent(connection, connection.connmark, data=b'', EOF=True)):
                        yield m
                    break
                except IOError:
                    for m in connection.reset():
                        yield m
                    break
                else:
                    for m in connection.write(ConnectionWriteEvent(connection, connection.connmark, data=data, EOF=False)):
                        yield m
        finally:
            connection.outputstream.close(connection.scheduler)
    def reconnect_init(self, connection):
        connection.inputstream = Stream()
        connection.outputstream = Stream(writebufferlimit=(self.writebufferlimit if self.buffering else 0),
                                         splitsize=self.splitsize)
        connection.subroutine(self._raw_writer(connection), False, '_raw_writer')
        for m in connection.waitForSend(RawConnectionStateEvent(RawConnectionStateEvent.CONNECTION_UP, connection, connection.connmark, self)):
            yield m
    def notconnected(self, connection):
        for m in Protocol.notconnected(self, connection):
            yield m
        for m in connection.waitForSend(RawConnectionStateEvent(RawConnectionStateEvent.CONNECTION_NOTCONNECTED, connection, connection.connmark, self)):
            yield m
    def closed(self, connection):
        for m in Protocol.closed(self, connection):
            yield m
        for m in connection.inputstream.write(b'', connection, True, True):
            yield m
        connection.terminate(connection._raw_writer)
        connection._raw_writer = None
        for m in connection.waitForSend(RawConnectionStateEvent(RawConnectionStateEvent.CONNECTION_DOWN, connection, connection.connmark, self)):
            yield m
    def error(self, connection):
        for m in Protocol.error(self, connection):
            yield m
        for m in connection.inputstream.error(connection, True):
            yield m
        connection.terminate(connection._raw_writer)
        connection._raw_writer = None
        for m in connection.waitForSend(RawConnectionStateEvent(RawConnectionStateEvent.CONNECTION_DOWN, connection, connection.connmark, self)):
            yield m
    def statematcher(self, connection, state = RawConnectionStateEvent.CONNECTION_DOWN, currentconn = True):
        if currentconn:
            return RawConnectionStateEvent.createMatcher(state, connection, connection.connmark)
        else:
            return RawConnectionStateEvent.createMatcher(state, connection)
    def client_connect(self, container, url, *args, **kwargs):
        '''
        Create a connection with raw protocol
        
        :param container: current routine container
        
        :param url: url to connect to (see Client)
        
        :param \*args, \*\*kwargs: other parameters to create a Client (except url, protocol and scheduler)
        
        :returns: `(connection, inputstream, outputstream)` where client is the created connection, inputstream
                  is the stream to read from the socket, outputstream is the stream to write to socket
        '''
        c = Client(url, self, container.scheduler, *args, **kwargs)
        c.start()
        yield (self.statematcher(c, RawConnectionStateEvent.CONNECTION_UP, False), self.statematcher(c, RawConnectionStateEvent.CONNECTION_NOTCONNECTED, False))
        if self.event.state == RawConnectionStateEvent.CONNECTION_UP:
            container.retval = (c, self.event.inputstream, self.event.outputstream)
        else:
            raise IOError('Connection failed')
    def redirect_outputstream(self, connection, stream):
        "Close current outputstream and output from the new stream"
        if not connection.connected:
            raise IOError('Connection is closed')
        if connection._raw_writer:
            connection.terminate(connection._raw_writer)
            del connection._raw_writer
        connection.outputstream = stream
        connection.subroutine(self._raw_writer(connection), False, '_raw_writer')
    def parse(self, connection, data, laststart):
        if connection.inputstream and not connection.inputstream.closed:
            if len(data) == laststart:
                return ([StreamDataEvent(connection.inputstream, StreamDataEvent.STREAM_EOF, data=b'')], 0)
            else:
                return ([StreamDataEvent(connection.inputstream, StreamDataEvent.STREAM_DATA, data=_copy(data))], 0)
        else:
            return ([], 0)



'''
Created on 2016/2/19

:author: hubo
'''

from vlcp.config import defaultconfig
from vlcp.protocol.jsonrpc import JsonRPC, JsonRPCRequestEvent
from vlcp.event.connection import ConnectionResetException

@defaultconfig
class OVSDB(JsonRPC):
    '''
    OVSDB protocol, this is a specialized JSON-RPC 1.0 protocol
    '''
    _default_defaultport = 6632
    # Only accept "echo" requests
    _default_allowedrequests = ('echo',)
    # Send "echo" requests when connection is idle
    _default_keepalivetime = 10
    # Disconnect when reply is not received for "echo" requests
    _default_keepalivetimeout = 3
    _default_tcp_nodelay = True
    def _respond_echo(self, connection):
        try:
            request_matcher = JsonRPCRequestEvent.createMatcher('echo', connection)
            connstate = self.statematcher(connection)
            while True:
                yield (request_matcher, connstate)
                if connection.matcher is request_matcher:
                    connection.event.canignore = True
                    reply = self.formatreply(connection.event.params, connection.event.id, connection)
                    for m in connection.write(reply, False):
                        yield m
                else:
                    break
        except ConnectionResetException:
            pass
    def reconnect_init(self, connection):
        for m in JsonRPC.reconnect_init(self, connection):
            yield m
        connection.subroutine(self._respond_echo(connection))
    def keepalive(self, connection):
        try:
            for m in connection.executeWithTimeout(self.keepalivetimeout, self.querywithreply('echo', [], connection, connection)):
                yield m
            if connection.timeout:
                for m in connection.reset(True):
                    yield m
        except Exception:
            for m in connection.reset(True):
                yield m



'''
Created on 2015/6/16

:author: hubo
'''
from __future__ import print_function, absolute_import, division 
import sys
from .core import QuitException, TimerEvent, SystemControlEvent
from .event import Event, withIndices
from contextlib import closing

class EventHandler(object):
    '''
    Runnable with an event handler model. 
    '''
    def __init__(self, scheduler = None, daemon = False):
        self.handlers = dict()
        self.scheduler = scheduler
        self.daemon = daemon
        self.registered = False
    def bind(self, scheduler):
        self.scheduler = scheduler
    def __iter__(self):
        '''
        Keep it like a iterator. Not very useful.
        '''
        return self
    def next(self):
        '''
        Keep it like a iterator. Not very useful.
        '''
        self.send(None)
    def __next__(self):
        '''
        Python 3 next
        '''
        self.send(None)
    def send(self, etup):
        '''
        Handle events
        '''
        return self.handlers[etup[1]](etup[0], self.scheduler)
    def _setDaemon(self):
        if not self.registered:
            self.registered = True
            self.scheduler.setDaemon(self, self.daemon)
    def registerHandler(self, matcher, handler):
        '''
        Register self to scheduler
        '''
        self.handlers[matcher] = handler
        self.scheduler.register((matcher,), self)
        self._setDaemon()
    def unregisterHandler(self, matcher):
        self.scheduler.unregister((matcher,), self)
        del self.handlers[matcher]
    def unregisterAllHandlers(self):
        self.scheduler.unregister(tuple(self.handlers.keys()), self)
        self.handlers.clear()
    def registerAllHandlers(self, handlerDict):
        '''
        Register self to scheduler
        '''
        self.handlers.update(handlerDict)
        if hasattr(handlerDict, 'keys'):
            self.scheduler.register(handlerDict.keys(), self)
        else:
            self.scheduler.register(tuple(h[0] for h in handlerDict), self)
        self._setDaemon()
    def close(self):
        self.scheduler.unregisterall(self)
        self.registered = False
    def registerExceptionHandler(self, handler):
        self.exceptionHandler = handler
    def registerQuitHandler(self, handler):
        self.quitHandler = handler
    def throw(self, typ, val = None, tb = None):
        if val is None:
            if isinstance(typ, type):
                val = typ()
            else:
                val = typ
                typ = type(val)
        if isinstance(val, QuitException):
            self.quitHandler(self.scheduler)
        else:
            self.exceptionHandler(val, self.scheduler)
    def exceptionHandler(self, val, scheduler):
        raise val
    def quitHandler(self, scheduler):
        raise StopIteration

@withIndices('type', 'routine')
class RoutineControlEvent(Event):
    canignore = False
    ASYNC_START = 'asyncstart'
    DELEGATE_FINISHED = 'delegatefinished'
    WAIT = 'wait'

class IllegalMatchersException(Exception):
    pass

class generatorwrapper(object):
    '''
    Default __repr__ of a generator is not readable, use a wrapper to improve the readability
    '''
    def __init__(self, run, name = 'iterator', classname = 'routine'):
        self.run = run
        self.name = name
        self.classname = classname
    def __iter__(self):
        return self.run
    def next(self):
        try:
            return next(self.run)
        except StopIteration:
            raise StopIteration
    def __next__(self):
        try:
            return next(self.run)
        except StopIteration:
            raise StopIteration
    def send(self, arg):
        try:
            return self.run.send(arg)
        except StopIteration:
            raise StopIteration
    def throw(self, typ, val = None, tb = None):
        try:
            return self.run.throw(typ, val, tb)
        except StopIteration:
            raise StopIteration
    def __repr__(self, *args, **kwargs):
        try:
            iterator = self.run.gi_frame.f_locals[self.name]
            try:
                return '<%s %r of %r at 0x%016X>' % (self.classname, iterator,
                                                       iterator.gi_frame.f_locals['self'],
                                                       id(iterator))
            except:
                return '<%s %r at 0x%016X>' % (self.classname, iterator, id(iterator))
        except:
            return repr(self.run)
    def close(self):
        return self.run.close()

def Routine(iterator, scheduler, asyncStart = True, container = None, manualStart = False, daemon = False):
    """
    This wraps a normal generator to become a VLCP routine. Usually you do not need to call this yourself;
    `container.start` and `container.subroutine` calls this automatically.
    """
    def run():
        iterself, re = yield
        rcMatcher = RoutineControlEvent.createMatcher(RoutineControlEvent.ASYNC_START, iterself)
        if manualStart:
            yield
        try:
            if asyncStart:
                scheduler.register((rcMatcher,), iterself)
                (event, m) = yield
                event.canignore = True
                scheduler.unregister((rcMatcher,), iterself)
            if container is not None:
                container.currentroutine = iterself
            if daemon:
                scheduler.setDaemon(iterself, True)
            matchers = next(iterator)
            try:
                scheduler.register(matchers, iterself)
            except:
                iterator.throw(IllegalMatchersException(matchers))
                raise
            while True:
                try:
                    etup = yield
                except GeneratorExit:
                    raise
                except:
                    #scheduler.unregister(matchers, iterself)
                    lmatchers = matchers
                    t,v,tr = sys.exc_info()  # @UnusedVariable
                    if container is not None:
                        container.currentroutine = iterself
                    matchers = iterator.throw(t,v)
                else:
                    #scheduler.unregister(matchers, iterself)
                    lmatchers = matchers
                    if container is not None:
                        container.event = etup[0]
                        container.matcher = etup[1]
                    if container is not None:
                        container.currentroutine = iterself
                    matchers = iterator.send(etup)
                try:
                    scheduler.unregister(set(lmatchers).difference(matchers), iterself)
                    scheduler.register(set(matchers).difference(lmatchers), iterself)
                except:
                    iterator.throw(IllegalMatchersException(matchers))
                    raise
        finally:
            if asyncStart:
                re.canignore = True
                scheduler.ignore(rcMatcher)
            # iterator.close() can be called in other routines, we should restore the currentroutine variable
            if container is not None:
                lastcurrentroutine = getattr(container, 'currentroutine', None)
                container.currentroutine = iterself
            else:
                lastcurrentroutine = None
            iterator.close()
            if container is not None:
                container.currentroutine = lastcurrentroutine
            scheduler.unregisterall(iterself)
    r = generatorwrapper(run())
    next(r)
    if asyncStart:
        re = RoutineControlEvent(RoutineControlEvent.ASYNC_START, r)
        r.send((r, re))
        waiter = scheduler.send(re)
        if waiter is not None:
            # This should not happen regularly
            def latencyStart(w):
                while w:
                    yield (w,)
                    w = scheduler.send(re)
            Routine(latencyStart(waiter), scheduler, False)
    else:
        r.send((r, None))
    return r

class RoutineException(Exception):
    """
    Special exception type raised from :py:method::`vlcp.event.runnable.RoutineContainer.withException`.
    e.matcher is set to the matcher and e.event is set to the matched event.
    """
    def __init__(self, matcher, event):
        Exception.__init__(self, matcher, event)
        self.matcher = matcher
        self.event = event

class MultipleException(Exception):
    """
    Special exception type raised from :py:method::`vlcp.event.runnable.RoutineContainer.executeAll`
    """
    def __init__(self, exceptions):
        Exception.__init__(self, '%d exceptions occurs in parallel execution' % (len(exceptions),) \
                           + ': ' + repr(exceptions[0]) + ', ...')
        self.exceptions = exceptions

class RoutineContainer(object):
    """
    A routine container groups several routines together and shares data among them. It is also
    used to pass important information like events, matchers and return values to the routine.
    
    Several attributes are commonly used:
    
    event
        When a routine returns from `yield`, `container.event` is set to the matched event object
        
    matcher
        When a routine returns from `yeild`, `container.event` is set to the matched event matcher
    
    timeout
        `waitWithTimeout`, `executeWithTimeout` uses this attribute to indicate whether the wait process
        exceeds the time limit
        
    retvalue
        many coroutine methods use this attribute to return a value
        
    currentroutine
        Always set to the executing routine - which is the wrapped routine object of the routine itself
        
    mainroutine
        Set to the main routine `container.main` (started by `container.start`)
        
    """
    def __init__(self, scheduler = None, daemon = False):
        """
        Create the routine container.
        
        :param scheduler: The scheduler. This must be set; if None is used, it must be set with
                          `container.bind(scheduler)` before using.
        
        :param daemon: If daemon = True, the main routine `container.main` is set to be a daemon routine.
                       A daemon routine does not stop the scheduler from quitting; if all non-daemon routines
                       are quit, the scheduler stops.
        """
        self.scheduler = scheduler
        self.daemon = daemon
    def bind(self, scheduler):
        """
        If scheduler is not specified, bind the scheduler
        """
        self.scheduler = scheduler
    def main(self):
        """
        The main routine method, should be rewritten to a generator method
        """
        raise NotImplementedError
    def start(self, asyncStart = False):
        """
        Start `container.main` as the main routine.
        
        :param asyncStart: if True, start the routine in background. By default, the routine
                           starts in foreground, which means it is executed to the first
                           `yield` statement before returning. If the started routine raises
                           an exception, the exception is re-raised to the caller of `start`
        """
        r = Routine(self.main(), self.scheduler, asyncStart, self, True, self.daemon)
        self.mainroutine = r
        try:
            next(r)
        except StopIteration:
            pass
        return r
    def subroutine(self, iterator, asyncStart = True, name = None, daemon = False):
        """
        Start extra routines in this container.
        
        :param iterator: A generator object i.e the return value of a generator method `my_routine()`
        
        :param asyncStart: if False, start the routine in foreground. By default, the routine
                           starts in background, which means it is not executed until the current caller
                           reaches the next `yield` statement or quit.
        
        :param name: if not None, `container.<name>` is set to the routine object. This is useful when
                     you want to terminate the routine from outside.
                     
        :param daemon: if True, this routine is set to be a daemon routine.
                       A daemon routine does not stop the scheduler from quitting; if all non-daemon routines
                       are quit, the scheduler stops. 
        """
        r = Routine(iterator, self.scheduler, asyncStart, self, True, daemon)
        if name is not None:
            setattr(self, name, r)
        # Call subroutine may change the currentroutine, we should restore it
        currentroutine = getattr(self, 'currentroutine', None)
        try:
            next(r)
        except StopIteration:
            pass
        self.currentroutine = currentroutine
        return r
    def terminate(self, routine = None):
        """
        Stop a routine.
        
        :param routine: if None, stop the main routine. If not None, it should be a routine object. You
                        can specify a name for a subroutine, and use `container.<name>` to retrieve it.
        """
        if routine is None:
            routine = self.mainroutine
        routine.close()
    def close(self):
        """
        Same as `terminate()`
        """
        self.terminate()
    def waitForSend(self, event):
        '''
        Send an event to the main event queue. Can call without delegate.
        '''
        waiter = self.scheduler.send(event)
        while waiter:
            yield (waiter,)
            waiter = self.scheduler.send(event)
    def waitWithTimeout(self, timeout, *matchers):
        """
        Wait for multiple event matchers, or until timeout. If time limit exceeds, `container.timeout` is
        set to True; otherwise it is set to False, and `event` `matcher` attributes are set.
        You should check the `timeout` attribute before accessing `event` or `matcher` attribute.
        
        :param timeout: a timeout value
        
        :param \*matchers: event matchers
        """
        if timeout is None:
            yield matchers
            self.timeout = False
        else:
            th = self.scheduler.setTimer(timeout)
            try:
                tm = TimerEvent.createMatcher(th)
                yield tuple(matchers) + (tm,)
                if self.matcher is tm:
                    self.timeout = True
                else:
                    self.timeout = False
            finally:
                self.scheduler.cancelTimer(th)
    def executeWithTimeout(self, timeout, subprocess):
        """
        Execute a subprocess with timeout. If time limit exceeds, the subprocess is terminated,
        and `container.timeout` is set to True; otherwise the `container.timeout` is set to False.
        
        You can uses `executeWithTimeout` with other help functions to create time limit for them::
        
            for m in container.executeWithTimeout(10, container.executeAll([routine1(), routine2()])):
                yield m
        """
        if timeout is None:
            for m in subprocess:
                yield m
            self.timeout = False
        else:
            th = self.scheduler.setTimer(timeout)
            try:
                tm = TimerEvent.createMatcher(th)
                try:
                    for m in self.withException(subprocess, tm):
                        yield m
                    self.timeout = False
                except RoutineException as exc:
                    if exc.matcher is tm:
                        self.timeout = True
                    else:
                        raise
            finally:
                self.scheduler.cancelTimer(th)
                subprocess.close()
    def doEvents(self):
        '''
        Suspend this routine until the next polling. This can be used to give CPU time for other routines and
        socket processings in long calculating procedures. Can call without delegate.
        '''
        self.scheduler.wantContinue()
        cm = SystemControlEvent.createMatcher(SystemControlEvent.CONTINUE)
        yield (cm,)
    def withException(self, subprocess, *matchers):
        """
        Monitoring event matchers while executing a subprocess. If events are matched before the subprocess ends,
        the subprocess is terminated and a RoutineException is raised.
        """
        try:
            for m in subprocess:
                yield tuple(m) + tuple(matchers)
                if self.matcher in matchers:
                    raise RoutineException(self.matcher, self.event)
        finally:
            subprocess.close()
    def withCallback(self, subprocess, callback, *matchers):
        """
        Monitoring event matchers while executing a subprocess. `callback(event, matcher)` is called each time
        an event is matched by any event matchers. If the callback raises an exception, the subprocess is terminated.
        """
        try:
            for m in subprocess:
                while True:
                    yield tuple(m) + tuple(matchers)
                    if self.matcher in matchers:
                        callback(self.event, self.matcher)
                    else:
                        break
        finally:
            subprocess.close()
                
    def waitForEmpty(self, queue):
        '''
        Wait for a queue to be empty. Can call without delegate
        '''
        while True:
            m = queue.waitForEmpty()
            if m is None:
                break
            else:
                yield (m,)
    def waitForAll(self, *matchers):
        """
        Wait until each matcher matches an event. When this coroutine method returns,
        `container.eventlist` is set to the list of events in the arriving order (may not
        be the same as the matchers); `container.eventdict` is set to a dictionary
        `{matcher1: event1, matcher2: event2, ...}`
        """
        eventdict = {}
        eventlist = []
        self.scheduler.register(matchers, self.currentroutine)
        try:
            ms = len(matchers)
            while ms:
                yield ()
                self.scheduler.unregister((self.matcher,), self.currentroutine)
                ms -= 1
                eventlist.append(self.event)
                eventdict[self.matcher] = self.event
            self.eventlist = eventlist
            self.eventdict = eventdict
        except:
            self.scheduler.unregister(matchers, self.currentroutine)
            raise
    def waitForAllToProcess(self, *matchers):
        """
        Similar to `waitForAll`, but set `canignore=True` for these events. This ensures
        blocking events are processed correctly.
        """
        eventdict = {}
        eventlist = []
        self.scheduler.register(matchers, self.currentroutine)
        try:
            ms = len(matchers)
            while ms:
                yield ()
                self.event.canignore = True
                self.scheduler.unregister((self.matcher,), self.currentroutine)
                ms -= 1
                eventlist.append(self.event)
                eventdict[self.matcher] = self.event
            self.eventlist = eventlist
            self.eventdict = eventdict
        except:
            self.scheduler.unregister(matchers, self.currentroutine)
            raise
    def waitForAllEmpty(self, *queues):
        """
        Wait for multiple queues to be empty at the same time.
        """
        matchers = [m for m in (q.waitForEmpty() for q in queues) if m is not None]
        while matchers:
            for m in self.waitForAll(*matchers):
                yield m
            matchers = [m for m in (q.waitForEmpty() for q in queues) if m is not None]
    def syscall_noreturn(self, func):
        '''
        Call a syscall method. A syscall method is executed outside of any routines, directly
        in the scheduler loop, which gives it chances to directly operate the event loop.
        See :py:method::`vlcp.event.core.Scheduler.syscall`.
        
        Can call without delegate.
        '''
        matcher = self.scheduler.syscall(func)
        while not matcher:
            e = RoutineControlEvent(RoutineControlEvent.WAIT, self.currentroutine)
            e.canignore = True
            for m in self.waitForSend(e):
                yield m
            yield (RoutineControlEvent.createMatcher(RoutineControlEvent.WAIT, self.currentroutine),)
            matcher = self.scheduler.syscall(func)
        yield (matcher,)
    def syscall(self, func, ignoreException = False):
        """
        Call a syscall method and retrieve its return value from `container.retvalue`
        """
        for m in self.syscall_noreturn(func):
            yield m
        if hasattr(self.event, 'exception'):
            raise self.event.exception[1]
        else:
            self.retvalue = self.event.retvalue
    def delegate(self, subprocess, forceclose = False):
        '''
        Run a subprocess without container support
        
        Many subprocess assume itself running in a specified container, it uses container reference
        like self.events. Calling the subprocess in other containers will fail.
        
        With delegate, you can call a subprocess in any container (or without a container)::
        
            for m in c.delegate(c.someprocess()):
                yield m
                
        Use `delegateOther` if you need to retrieve the return value.
        '''
        def delegateroutine():
            try:
                for m in subprocess:
                    yield m
            except:
                typ, val, tb = sys.exc_info()
                e = RoutineControlEvent(RoutineControlEvent.DELEGATE_FINISHED, self.currentroutine)
                e.canignore = True
                for m in self.waitForSend(e):
                    yield m
                raise
            else:
                e = RoutineControlEvent(RoutineControlEvent.DELEGATE_FINISHED, self.currentroutine)
                e.canignore = True
                for m in self.waitForSend(e):
                    yield m
        r = self.subroutine(generatorwrapper(delegateroutine(), 'subprocess', 'delegate'), True)
        finish = RoutineControlEvent.createMatcher(RoutineControlEvent.DELEGATE_FINISHED, r)
        # As long as we do not use self.event to read the event, we are safe to receive them from other containers
        try:
            yield (finish,)
        finally:
            if forceclose:
                r.close()
    def beginDelegateOther(self, subprocess, container, retnames = ('retvalue',)):
        '''
        Start the delegate routine, but do not wait for result, instead returns a matcher in self.retvalue.
        Useful for advanced delegates (e.g. delegate multiple subprocesses in the same time).
        This is NOT a coroutine method.
        
        :param subprocess: a subroutine
        
        :param container: container in which to start the routine
        
        :param retnames: get return values from keys
        
        :returns: (matcher, routine) where matcher is a event matcher to get the delegate result, routine is the created routine
        '''
        def delegateroutine():
            try:
                for m in subprocess:
                    yield m
            except:
                typ, val, tb = sys.exc_info()
                e = RoutineControlEvent(RoutineControlEvent.DELEGATE_FINISHED, container.currentroutine, exception = val)
                e.canignore = True
                for m in container.waitForSend(e):
                    yield m
                raise val
            else:
                e = RoutineControlEvent(RoutineControlEvent.DELEGATE_FINISHED, container.currentroutine,
                                        result = tuple(getattr(container, n, None) for n in retnames))
                e.canignore = True
                for m in container.waitForSend(e):
                    yield m
        r = container.subroutine(generatorwrapper(delegateroutine(), 'subprocess', 'delegate'), True)
        return (RoutineControlEvent.createMatcher(RoutineControlEvent.DELEGATE_FINISHED, r), r)
    def delegateOther(self, subprocess, container, retnames = ('retvalue',), forceclose = False):
        '''
        Another format of delegate allows delegate a subprocess in another container, and get some returning values
        the subprocess is actually running in 'container'. ::
        
            for m in self.delegateOther(c.method(), c):
                yield m
            ret = self.retvalue
        '''
        finish, r = self.beginDelegateOther(subprocess, container, retnames)
        try:
            yield (finish,)
            if hasattr(self.event, 'exception'):
                raise self.event.exception
            for n, v in zip(retnames, self.event.result):
                setattr(self, n, v)
        finally:
            if forceclose:
                r.close()
    def executeAll(self, subprocesses, container = None, retnames = ('retvalue',), forceclose = True):
        '''
        Execute all subprocesses and get the return values. Return values are in self.retvalue.
        
        :param subprocesses: sequence of subroutines (generators)
        
        :param container: if specified, run subprocesses in another container.
        
        :param retnames: get return value from container.(name) for each name in retnames.
        
        :param forceclose: force close the routines on exit, so all the subprocesses are terminated
                           on timeout if used with executeWithTimeout
        
        :returns: a list of tuples, one for each subprocess, with value of retnames inside:
                  `[('retvalue1',),('retvalue2',),...]`
        '''
        if not subprocesses:
            self.retvalue = []
            return
        subprocesses = list(subprocesses)
        if len(subprocesses) == 1 and (container is None or container is self) and forceclose:
            # Directly run the process to improve performance
            with closing(subprocesses[0]) as sp:
                for m in sp:
                    yield m
            self.retvalue = [tuple(getattr(self, n, None) for n in retnames)]
            return
        if container is None:
            container = self
        delegates = [self.beginDelegateOther(p, container, retnames) for p in subprocesses]
        matchers = [d[0] for d in delegates]
        try:
            for m in self.waitForAll(*matchers):
                yield m
            events = [self.eventdict[m] for m in matchers]
            exceptions = [e.exception for e in events if hasattr(e, 'exception')]
            if exceptions:
                if len(exceptions) == 1:
                    raise exceptions[0]
                else:
                    raise MultipleException(exceptions)
            self.retvalue = [e.result for e in events]
        finally:
            if forceclose:
                for d in delegates:
                    try:
                        container.terminate(d[1])
                    except:
                        pass



import itertools
import os

import vlcp.service.sdn.ofpportmanager as ofpportmanager
import vlcp.service.kvdb.objectdb as objectdb
import vlcp.service.sdn.ioprocessing as iop
from vlcp.service.sdn.flowbase import FlowBase
from vlcp.server.module import depend, callAPI
from vlcp.config.config import defaultconfig
from vlcp.event.runnable import RoutineContainer
from vlcp.service.sdn.ofpmanager import FlowInitialize
from vlcp.utils.ethernet import mac_addr_bytes, ip4_addr_bytes,ip4_icmp_payload,\
    ethernet_l7, ip4_packet_l7, ip4_payload,ICMP_ECHOREPLY,icmp_bestparse,icmp_echo
from vlcp.utils.flowupdater import FlowUpdater
from vlcp.protocol.openflow.openflow import OpenflowConnectionStateEvent, OpenflowAsyncMessageEvent
from vlcp.utils.networkmodel import SubNet,RouterPort
from namedstruct.stdprim import uint16

class ICMPResponderUpdater(FlowUpdater):
    def __init__(self,connection,parent):
        super(ICMPResponderUpdater,self).__init__(connection,(),('icmpresponderupdate',connection),parent._logger)
        self.parent = parent
        self._lastlognets = ()
        self._lastlogports = ()
        self._lastsubnetsinfo = dict()
        self._orig_initialkeys = ()

    def main(self):
        try:
            self.subroutine(self._update_handler(),True,"update_handler_routine")

            # use controller to reply icmp ping ,so start routine handler packet in
            if not self.parent.prepush:
                self.subroutine(self._icmp_packetin_handler(),True,"icmp_packetin_handler_routine")

            for m in FlowUpdater.main(self):
                yield m
        finally:
            if hasattr(self,"update_handler_routine"):
                self.update_handler_routine.close()

            if hasattr(self,"icmp_packetin_handler_routine"):
                self.icmp_packetin_handler_routine.close()

    def _icmp_packetin_handler(self):
        conn = self._connection
        ofdef = self._connection.openflowdef
        l3input = self.parent._gettableindex("l3input",self._connection.protocol.vhost)
        
        transactid = uint16.create(os.urandom(2)) 

        def send_packet_out(portid,packet):
            for m in self.execute_commands(conn,
                        [
                            ofdef.ofp_packet_out(
                                buffer_id = ofdef.OFP_NO_BUFFER,
                                in_port = ofdef.OFPP_CONTROLLER,
                                actions = [
                                    ofdef.ofp_action_output(port = portid,
                                                            max_len = ofdef.OFPCML_NO_BUFFER
                                                            )
                                ],
                                data = packet._tobytes()
                            )
                        ]):
                yield m

        icmp_packetin_matcher = OpenflowAsyncMessageEvent.createMatcher(ofdef.OFPT_PACKET_IN,None,None,l3input,2,
                                                            self._connection,self._connection.connmark)

        while True:
            yield (icmp_packetin_matcher,)
            msg = self.event.message
            inport = ofdef.ofp_port_no.create(ofdef.get_oxm(msg.match.oxm_fields,ofdef.OXM_OF_IN_PORT))

            # it must be icmp packet ...
            icmp_packet = ethernet_l7.create(msg.data)
            
            transactid = (transactid + 1) & 0xffff
            reply_packet = ip4_packet_l7((ip4_payload,ip4_icmp_payload),
                                         (icmp_bestparse, icmp_echo),
                                        dl_src = icmp_packet.dl_dst,
                                        dl_dst = icmp_packet.dl_src,
                                        ip_src = icmp_packet.ip_dst,
                                        ip_dst = icmp_packet.ip_src,
                                        frag_off = icmp_packet.frag_off,
                                        ttl = 128,
                                        identifier = transactid,
                                        icmp_type = ICMP_ECHOREPLY,
                                        icmp_code = icmp_packet.icmp_code,
                                        icmp_id = icmp_packet.icmp_id,
                                        icmp_seq = icmp_packet.icmp_seq,
                                        data = icmp_packet.data
                                        )
           
            self.subroutine(send_packet_out(inport,reply_packet))
    def _update_handler(self):

        # when lgport,lgnet,phyport,phynet object change , receive this event from ioprocessing module
        dataobjectchange = iop.DataObjectChanged.createMatcher(None,None,self._connection)

        while True:
            yield (dataobjectchange,)

            # save to instance attr ,  us in other method
            self._lastlogports,_,self._lastlognets,_ = self.event.current
            self._update_walk()

    def _walk_lgport(self,key,value,walk,save):

        if value is not None:
            save(key)
            if hasattr(value,'subnet'):
                try:
                    subnetobj = walk(value.subnet.getkey())
                except KeyError:
                    pass
                else:
                    save(value.subnet.getkey())
                    if subnetobj is not None and hasattr(subnetobj,"router"):
                        try:
                            _ = walk(subnetobj.router.getkey())
                        except KeyError:
                            pass
                        else:
                            save(subnetobj.router.getkey())

    def _walk_lgnet(self,key,value,walk,save):
        save(key)
        # if value is None, also save its key
        # means watch key, when created , we will recv event

    def _update_walk(self):
        lgportkeys = [p.getkey() for p,_ in self._lastlogports]
        lgnetkeys = [p.getkey() for p,_ in self._lastlognets]

        self._initialkeys = lgportkeys + lgnetkeys
        self._orig_initialkeys = lgportkeys + lgnetkeys
        
        self._walkerdict = dict(itertools.chain(((p,self._walk_lgport) for p in lgportkeys),
                                                ((n,self._walk_lgnet) for n in lgnetkeys)))
        
        self.subroutine(self.restart_walk(),False)
    
    def reset_initialkeys(self,keys,values):
        # walk map  logicalport --> subnet ---> routerport
        # we get subnet object, add keys to initialkeys, 
        # when subnet update, it will restart walk ,, after we will get new routerport
        
        subnetkeys = [k for k,v in zip(keys,values) if v is not None and not v.isdeleted() and
                      v.isinstance(SubNet)]

        self._initialkeys = tuple(itertools.chain(self._orig_initialkeys,subnetkeys))
    
    def updateflow(self, connection, addvalues, removevalues, updatedvalues):
        try:
            allobjects = set(o for o in self._savedresult if o is not None and not o.isdeleted())

            lastsubnetsinfo = self._lastsubnetsinfo
            currentlognetsinfo = dict((n,id) for n,id in self._lastlognets if n in allobjects)
            currentrouterportsinfo = dict((o.subnet,o) for o in allobjects
                                            if o.isinstance(RouterPort))
            currentsubnetsinfo = dict((o,(getattr(currentrouterportsinfo[o],"ip_address",getattr(o,"gateway",None)),
                                          self.parent.inroutermac,o.network.id,currentlognetsinfo[o.network]))
                                        for o in allobjects if o.isinstance(SubNet)
                                            and hasattr(o,"router") and o in currentrouterportsinfo
                                            and o.network in currentlognetsinfo
                                            and (hasattr(currentrouterportsinfo[o],"ip_address")
                                                or hasattr(o,"gateway"))
                                            and ( not hasattr(o,"isexternal") or o.isexternal == False))
            self._lastsubnetsinfo = currentsubnetsinfo

            ofdef = connection.openflowdef
            vhost = connection.protocol.vhost
            l3input = self.parent._gettableindex("l3input",vhost)

            cmds = []

            if connection.protocol.disablenxext:
                def match_network(nid):
                    return ofdef.create_oxm(ofdef.OXM_OF_METADATA_W, (nid & 0xffff) << 32,
                                            b'\x00\x00\xff\xff\x00\x00\x00\x00')
            else:
                def match_network(nid):
                    return ofdef.create_oxm(ofdef.NXM_NX_REG4, nid)

            # prepush or not ,, it is same , so ..
            def _deleteicmpflows(ipaddress, macaddress, networkid):
                    return [
                        ofdef.ofp_flow_mod(
                            cookie = 0x2,
                            cookie_mask = 0xffffffffffffffff,
                            table_id = l3input,
                            command = ofdef.OFPFC_DELETE,
                            priority = ofdef.OFP_DEFAULT_PRIORITY + 1,
                            buffer_id = ofdef.OFP_NO_BUFFER,
                            out_port = ofdef.OFPP_ANY,
                            out_group = ofdef.OFPG_ANY,
                            match = ofdef.ofp_match_oxm(
                                oxm_fields = [
                                    ofdef.create_oxm(ofdef.NXM_NX_REG4,networkid),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST,mac_addr_bytes(macaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE,ofdef.ETHERTYPE_IP),
                                    ofdef.create_oxm(ofdef.OXM_OF_IPV4_DST,ip4_addr_bytes(ipaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_IP_PROTO,ofdef.IPPROTO_ICMP),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_TYPE,8),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_CODE,0)
                                ]
                            )
                        )
                    ]

            if not self.parent.prepush:
                def _createicmpflows(ipaddress, macaddress, networkid):
                    return [
                        ofdef.ofp_flow_mod(
                            cookie = 0x2,
                            cookie_mask = 0xffffffffffffffff,
                            table_id = l3input,
                            command = ofdef.OFPFC_ADD,
                            # icmp to router matcher same as ip forward to router
                            # so priority + 1
                            priority = ofdef.OFP_DEFAULT_PRIORITY + 1,
                            buffer_id = ofdef.OFP_NO_BUFFER,
                            out_port = ofdef.OFPP_ANY,
                            out_group = ofdef.OFPG_ANY,
                            match = ofdef.ofp_match_oxm(
                                oxm_fields = [
                                    match_network(networkid),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST,mac_addr_bytes(macaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE,ofdef.ETHERTYPE_IP),
                                    ofdef.create_oxm(ofdef.OXM_OF_IPV4_DST,ip4_addr_bytes(ipaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_IP_PROTO,ofdef.IPPROTO_ICMP),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_TYPE,8),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_CODE,0)
                                ]
                            ),
                            instructions = [
                                ofdef.ofp_instruction_actions(
                                    actions = [
                                        ofdef.ofp_action_output(
                                            port = ofdef.OFPP_CONTROLLER,
                                            max_len = ofdef.OFPCML_NO_BUFFER
                                        )
                                    ]
                                )
                            ]
                        )
                    ]
            else:
                def _createicmpflows(ipaddress, macaddress, networkid):
                    return [
                        ofdef.ofp_flow_mod(
                            cookie = 0x2,
                            cookie_mask = 0xffffffffffffffff,
                            table_id = l3input,
                            command = ofdef.OFPFC_ADD,
                            # icmp to router matcher same as ip forward to router
                            # so priority + 1
                            priority = ofdef.OFP_DEFAULT_PRIORITY + 1,
                            buffer_id = ofdef.OFP_NO_BUFFER,
                            out_port = ofdef.OFPP_ANY,
                            out_group = ofdef.OFPG_ANY,
                            match = ofdef.ofp_match_oxm(
                                oxm_fields = [
                                    match_network(networkid),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST,mac_addr_bytes(macaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE,ofdef.ETHERTYPE_IP),
                                    ofdef.create_oxm(ofdef.OXM_OF_IPV4_DST,ip4_addr_bytes(ipaddress)),
                                    ofdef.create_oxm(ofdef.OXM_OF_IP_PROTO,ofdef.IPPROTO_ICMP),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_TYPE,8),
                                    ofdef.create_oxm(ofdef.OXM_OF_ICMPV4_CODE,0)
                                ]
                            ),
                            instructions = [
                                ofdef.ofp_instruction_actions(
                                    actions = [
                                        ofdef.nx_action_reg_move(
                                            n_bits = 48,
                                            src = ofdef.OXM_OF_ETH_SRC,
                                            dst = ofdef.OXM_OF_ETH_DST
                                        ),
                                        ofdef.ofp_action_set_field(
                                            field = ofdef.create_oxm(
                                                ofdef.OXM_OF_ETH_SRC,
                                                ofdef.mac_addr(macaddress)
                                            )
                                        ),
                                        ofdef.nx_action_reg_move(
                                            n_bits = 32,
                                            src = ofdef.OXM_OF_IPV4_SRC,
                                            dst = ofdef.OXM_OF_IPV4_DST
                                        ),
                                        ofdef.ofp_action_set_field(
                                            field = ofdef.create_oxm(
                                                ofdef.OXM_OF_IPV4_SRC,
                                                ofdef.ip4_addr(ipaddress)
                                            )
                                        ),
                                        ofdef.ofp_action_set_field(
                                            field = ofdef.create_oxm(
                                                ofdef.OXM_OF_ICMPV4_TYPE,
                                                ICMP_ECHOREPLY
                                            )
                                        ),
                                        ofdef.ofp_action_nw_ttl(
                                            nw_ttl = 128
                                        ),
                                        ofdef.ofp_action_output(
                                            port = ofdef.OFPP_IN_PORT
                                        )
                                    ]
                                )
                            ]
                        )
                    ]

            for subnet in lastsubnetsinfo.keys():
                if subnet not in currentsubnetsinfo\
                        or (subnet in currentsubnetsinfo and lastsubnetsinfo[subnet] != currentsubnetsinfo[subnet]):
                    # subnet remove  or subnet info changed , remove flow info
                    ip_address, mac_address, networkid, nid = lastsubnetsinfo[subnet]

                    remove_arp = {(ip_address,mac_address,networkid,True),}
                    for m in callAPI(self, 'arpresponder', 'removeproxyarp', {'connection':connection,
                                                                              'arpentries': remove_arp}):
                        yield m

                    cmds.extend(_deleteicmpflows(ip_address,mac_address,nid))

            for m in self.execute_commands(connection, cmds):
                yield m

            for subnet in currentsubnetsinfo.keys():
                if subnet not in lastsubnetsinfo\
                        or (subnet in lastsubnetsinfo and lastsubnetsinfo[subnet] != currentsubnetsinfo[subnet]):

                    ip_address, mac_address, networkid, nid = currentsubnetsinfo[subnet]

                    add_arp = {(ip_address,mac_address,networkid,True),}
                    for m in callAPI(self, 'arpresponder', 'createproxyarp', {'connection': connection,
                                                                               'arpentries': add_arp}):
                        yield m

                    cmds.extend(_createicmpflows(ip_address,mac_address,nid))

            for m in self.execute_commands(connection, cmds):
                yield m

        except Exception:
            self._logger.warning("Unexpected exception in icmp_flow_updater, ignore it! Continue",exc_info=True)

@defaultconfig
@depend(ofpportmanager.OpenflowPortManager,objectdb.ObjectDB)
class ICMPResponder(FlowBase):
    """
    Respond ICMP echo (ping) requests to the gateway
    """
    _tablerequest = (
        ("l3input",("l2input",),""),
        ("l2output",("l3input",),"")
    )
    # True : reply icmp ping with flow
    # False: reply icmp ping with controller PACKET_IN/PACKET_OUT
    #
    # Must use prepush=True with OpenvSwitch 2.5+
    #
    _default_prepush = False

    # "Gateway" responds with this MAC address
    _default_inroutermac = '1a:23:67:59:63:33'

    def __init__(self,server):
        super(ICMPResponder,self).__init__(server)
        self.app_routine = RoutineContainer(self.scheduler)
        self.app_routine.main = self._main
        self.routines.append(self.app_routine)
        self._flowupdater = dict()

    def _main(self):

        flowinit = FlowInitialize.createMatcher(_ismatch=lambda x: self.vhostbind is None or
                                                x.vhost in self.vhostbind)
        conndown = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_DOWN,
                                                _ismatch=lambda x:self.vhostbind is None or
                                                x.createby.vhost in self.vhostbind)
        while True:
            yield (flowinit,conndown)
            if self.app_routine.matcher is flowinit:
                c = self.app_routine.event.connection
                self.app_routine.subroutine(self._init_conn(c))
            if self.app_routine.matcher is conndown:
                c = self.app_routine.event.connection
                self.app_routine.subroutine(self._remove_conn(c))

    def _init_conn(self,conn):

        if conn in self._flowupdater:
            updater = self._flowupdater.pop(conn)
            updater.close()

        updater = ICMPResponderUpdater(conn,self)
        self._flowupdater[conn] = updater
        updater.start()

        if False:
            yield

    def _remove_conn(self,conn):

        if conn in self._flowupdater:
            updater = self._flowupdater.pop(conn)
            updater.close()

        if False:
            yield



'''
Created on 2015/6/17

:author: hubo
'''
from __future__ import print_function, absolute_import, division
import select
from vlcp.event import PollEvent, SystemControlLowPriorityEvent
import errno
import time
import sys
from vlcp.event.core import InterruptedBySignalException

if hasattr(select, 'epoll'):
    class EPollPolling(object):
        '''
        Poll event from epoll
        '''
        def __init__(self, options = select.EPOLLET, maxwait = 60):
            '''
            Constructor
            '''
            self.epoll = select.epoll()
            self.defaultoption = options
            self.mask = select.EPOLLIN|select.EPOLLPRI|select.EPOLLOUT|\
                select.EPOLLRDNORM|select.EPOLLRDBAND|select.EPOLLWRNORM|\
                select.EPOLLWRBAND|select.EPOLLONESHOT|select.EPOLLET
            self.socketCounter = 0
            self.maxwait = maxwait
            self.daemons = set()
            self.shouldraise = False
        def register(self, fd, options, daemon = False):
            if hasattr(fd, 'fileno'):
                fd = fd.fileno()
            self.epoll.register(fd, options & self.mask | self.defaultoption)
            if not daemon:
                self.socketCounter+=1
            else:
                self.daemons.add(fd)
        def unregister(self, fd, daemon = False):
            try:
                if hasattr(fd, 'fileno'):
                    fd = fd.fileno()
                self.epoll.unregister(fd)
                daemon = fd in self.daemons
                if not daemon:
                    self.socketCounter-=1
                else:
                    self.daemons.discard(fd)
            except IOError:
                return
            except ValueError:
                if not daemon:
                    self.socketCounter-=1
                return
        def modify(self, fd, options):
            self.epoll.modify(fd, options & self.mask | self.defaultoption)
        def setdaemon(self, fd, daemon):
            if hasattr(fd, 'fileno'):
                fd = fd.fileno()
            isdaemon = fd in self.daemons
            if isdaemon != daemon:
                if daemon:
                    self.daemons.add(fd)
                    self.socketCounter -= 1
                else:
                    self.daemons.discard(fd)
                    self.socketCounter += 1
        def pollEvents(self, wait):
            ret = []
            epwrite = select.EPOLLOUT|select.EPOLLWRNORM|select.EPOLLWRBAND
            epread = select.EPOLLIN|select.EPOLLRDNORM|select.EPOLLRDBAND|select.EPOLLPRI
            eperr = select.EPOLLERR
            ephup = select.EPOLLHUP
            if self.socketCounter <= 0 and wait is None:
                return ([], True)
            generateFree = False
            if wait is None or (self.maxwait is not None and wait > self.maxwait):
                generateFree = True
                wait = self.maxwait
            events = []
            try:
                interrupted = False
                self.shouldraise = True
                if wait is None:
                    events = self.epoll.poll()
                else:
                    events = self.epoll.poll(wait)
                self.shouldraise = False
            except InterruptedBySignalException:
                interrupted = True
            except IOError as exc:
                if exc.args[0] == errno.EINTR:
                    interrupted = True
                else:
                    raise
            finally:
                self.shouldraise = False                
            for fd, e in events:
                if e & epwrite:
                    ret.append(PollEvent(fd, PollEvent.WRITE_READY, e & epwrite))
                if e & epread:
                    ret.append(PollEvent(fd, PollEvent.READ_READY, e & epread))
                if e & eperr:
                    ret.append(PollEvent(fd, PollEvent.ERROR, e & eperr))
                if e & ephup:
                    ret.append(PollEvent(fd, PollEvent.HANGUP, e & ephup))
            if not ret and generateFree and not interrupted:
                ret.append(SystemControlLowPriorityEvent(SystemControlLowPriorityEvent.FREE))
            return (ret, False)
        def onmatch(self, fileno, category, add):
            pass

class SelectPolling(object):
    '''
    Compatible event polling with select
    '''
    def __init__(self, options = 0, maxwait = 60):
        '''
        Constructor
        '''
        self.readfiles = set()
        self.writefiles = set()
        self.errorfiles = set()
        self.daemons = set()
        self.socketCounter = 0
        self.maxwait = maxwait
        self.shouldraise = False
    def register(self, fd, options, daemon = False):
        if hasattr(fd, 'fileno'):
            fd = fd.fileno()
        self.errorfiles.add(fd)
        if not daemon:
            self.socketCounter+=1
        else:
            self.daemons.add(fd)
    def unregister(self, fd, daemon = False):
        try:
            if hasattr(fd, 'fileno'):
                fd = fd.fileno()
            daemon = fd in self.daemons
            if not daemon:
                self.socketCounter-=1
            else:
                self.daemons.discard(fd)
            if hasattr(fd, 'fileno'):
                fd = fd.fileno()
            self.errorfiles.discard(fd)
            #self.readfiles.discard(fd)
            #self.writefiles.discard(fd)
        except IOError:
            return
    def modify(self, fd, options):
        pass
    def setdaemon(self, fd, daemon):
        if hasattr(fd, 'fileno'):
            fd = fd.fileno()
        isdaemon = fd in self.daemons
        if isdaemon != daemon:
            if daemon:
                self.daemons.add(fd)
                self.socketCounter -= 1
            else:
                self.daemons.discard(fd)
                self.socketCounter += 1
    def pollEvents(self, wait):
        ret = []
        if self.socketCounter <= 0 and wait is None:
            return ([], True)
        generateFree = False
        if wait is None or (self.maxwait is not None and wait > self.maxwait):
            generateFree = True
            wait = self.maxwait
        events = [[], [], []]
        try:
            interrupted = False
            self.shouldraise = True
            if wait is None:
                events = select.select(self.readfiles, self.writefiles, self.errorfiles)
            elif not self.readfiles and not self.writefiles and not self.errorfiles:
                time.sleep(wait)
            else:
                events = select.select(self.readfiles, self.writefiles, self.errorfiles, wait)
            self.shouldraise = False
        except InterruptedBySignalException:
            interrupted = True
        except IOError as exc:
            if exc.args[0] == errno.EINTR:
                interrupted = True
            else:
                raise IOError('Some of the fds are invalid, maybe some sockets are not unregistered', exc)        
        finally:
            self.shouldraise = False
        for fd in events[0]:
            ret.append(PollEvent(fd, PollEvent.READ_READY, PollEvent.READ_READY))
        for fd in events[1]:
            ret.append(PollEvent(fd, PollEvent.WRITE_READY, PollEvent.WRITE_READY))
        for fd in events[2]:
            ret.append(PollEvent(fd, PollEvent.ERROR, PollEvent.ERROR))
        if not ret and generateFree and not interrupted:
            ret.append(SystemControlLowPriorityEvent(SystemControlLowPriorityEvent.FREE))
        return (ret, False)
    def onmatch(self, fileno, category, add):
        if add:
            if category is None or category == PollEvent.READ_READY:
                self.readfiles.add(fileno)
            if category is None or category == PollEvent.WRITE_READY:
                self.writefiles.add(fileno)
        else:
            if category is None or category == PollEvent.READ_READY:
                self.readfiles.discard(fileno)
            if category is None or category == PollEvent.WRITE_READY:
                self.writefiles.discard(fileno)            
